local addonName, ns = ...

local _G = _G
local format = string.format
local IsUsableAction = IsUsableAction
local IsActionInRange = IsActionInRange
local CreateFrame = CreateFrame
local replace = string.gsub
local UnitClass = UnitClass
local match = string.match
local tput = table.insert
local lower = string.lower
local find = string.find

local bBars = CreateFrame('frame', 'bBars', UIParent)
bBars:SetAllPoints()

local IsPandaria = select(4, GetBuildInfo()) > 40300
local class = select(2, UnitClass('player'))
local pname = select(1, UnitName('player'))

local bBars_config = false
local bBars_frames = {}

local statusbartex = 'Interface\\Buttons\\WHITE8x8'
local backgroundtex = 'Interface\\Buttons\\WHITE8x8'
local bBars_font = 'Interface\\AddOns\\'..addonName..'\\font1.ttf' -- select(1, ChatFontNormal:GetFont())

local fontStyle = 'THINOUTLINEMONOCHROME'
local fontSize = 10
local bBars_warning = bBars:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
bBars_warning:SetFont(select(1, ChatFontNormal:GetFont()), 62, 'OUTLINE')
bBars_warning:SetText('|cff00ddaashift + mousewheel to scale bar|nalt + mousewheel to change buttons padding|nmousewheel to change columns|r')
bBars_warning:SetPoint('CENTER')
bBars_warning:SetAlpha(.3)
bBars_warning:Hide()


local petbattle = ''
if IsPandaria then petbattle = '[petbattle]' end

local nullfunc = function() end

local scale = max(0.64, min(1.15, GetCVar('uiScale') or UIParent:GetScale() or 768/GetScreenHeight()))
local mult = 768/match(GetCVar('gxResolution'), '%d+x(%d+)') / scale
local Fix = function(v) return mult*math.floor(v/mult+.5) end
local round = function(number, decimals)
	if not decimals then decimals = 0 end
    return (("%%.%df"):format(decimals)):format(number)
end

local bBars_default_positions = {
	['bExpBarFrame'] = 					{ a = 'BOTTOM', p = 'UIParent', pa = 'BOTTOM', x = 0, y = 0, col = 12, hide = 0, size = 200 },
	['bRepBarFrame'] = 					{ a = 'BOTTOM', p = 'UIParent', pa = 'BOTTOM', x = 0, y = 10, col = 12, hide = 0, size = 200 },
	['bPetBattleFrame'] = 				{ a = 'BOTTOM', p = 'UIParent', pa = 'BOTTOM', x = 0, y = 0, col = 12, hide = 0, gap = 3 },
	['bBattlePetExpBarFrame'] = 		{ a = 'BOTTOM', p = 'bPetBattleFrame', pa = 'TOP', x = 0, y = 0, col = 12, hide = 0, size = 200 },
	['bMultiBar1'] = 					{ a = 'BOTTOM', p = 'bRepBarFrame', pa = 'TOP', x = 0, y = 0, col = 12, hide = 0, gap = 3 },
	['bMultiBar2'] = 					{ a = 'CENTER', p = 'bBars', pa = 'CENTER', x = 0, y = 0, col = 12, hide = 1, gap = 3 },
	['bMultiBar3'] = 					{ a = 'RIGHT', p = 'bBars', pa = 'RIGHT', x = 0, y = 0, col = 1, hide = 0, gap = 3 },
	['bMultiBar4'] = 					{ a = 'RIGHT', p = 'bMultiBar3', pa = 'LEFT', x = 0, y = 0, col = 1, hide = 0, gap = 3 },
	['bMultiBar5'] = 					{ a = 'BOTTOM', p = 'bMultiBar1', pa = 'TOP', x = 0, y = 0, col = 12, hide = 0, gap = 3 },
	['bMultiBar6'] = 					{ a = 'BOTTOM', p = 'bMultiBar5', pa = 'TOP', x = 0, y = 0, col = 12, hide = 0, gap = 3 },
	['bMultiBar7'] = 					{ a = 'TOP', p = 'bMultiBar2', pa = 'BOTTOM', x = 0, y = 0, col = 12, hide = 1, gap = 3 },
	['bMultiBar8'] = 					{ a = 'TOP', p = 'bMultiBar7', pa = 'BOTTOM', x = 0, y = 0, col = 12, hide = 1, gap = 3 },
	['bWatchFrame'] = 					{ a = 'RIGHT', p = 'bMultiBar4', pa = 'LEFT', x = 0, y = 0, col = 1, hide = 0, size = 25, gap = 3 },
	['bShapeshiftFrame'] = 				{ a = 'BOTTOMLEFT', p = 'bMultiBar6', pa = 'TOPLEFT', x = 0, y = 0, col = 12, hide = 0, size = 25, gap = 3 },
	['MultiCastActionBarFrame'] = 		{ a = 'BOTTOMLEFT', p = 'bMultiBar6', pa = 'TOPLEFT', x = 0, y = 0, col = 12, hide = 0, size = 30, gap = 3 },
	
	['bPetFrame'] = 					{ a = 'BOTTOMLEFT', p = 'bMultiBar6', pa = 'TOPLEFT', x = 0, y = 0, col = 12, hide = 0, size = 25, gap = 3 },
	['MultiCastActionBarFrame'] = 		{ a = 'BOTTOM', p = 'bMultiBar6', pa = 'TOP', x = 0, y = 0, col = 12, hide = 0 },
	['bMicroMenuFrame'] = 				{ a = 'LEFT', p = 'bMultiBar1', pa = 'RIGHT', x = 0, y = 0, col = 12, hide = 0, size = 20, gap = 3, col = 6 },
	['bBagsFrame'] = 					{ a = 'BOTTOMLEFT', p = 'bMicroMenuFrame', pa = 'BOTTOMRIGHT', x = 0, y = 0, col = 12, hide = 1 },
	['bVehicleFrame'] = 				{ a = 'BOTTOMLEFT', p = 'bMicroMenuFrame', pa = 'TOPLEFT', x = 0, y = 0, col = 12, hide = 0 },
	['bExtraBarFrame'] = 				{ a = 'BOTTOM', p = 'bVehicleFrame', pa = 'TOP', x = 0, y = 0, col = 12, hide = 0 },
}

if LibStub == nil then
	LibStub = _G.LibStub or {libs = {}, minors = {}, minor = 2}
	function LibStub:NewLibrary(major, minor)
		if self.minors[major] and self.minors[major] >= minor then return nil end
		self.minors[major], self.libs[major] = minor, self.libs[major] or {}
		return self.libs[major], self.minors[major]
	end
	function LibStub:GetLibrary(major, silent) return self.libs[major], self.minors[major] end
	function LibStub:IterateLibraries() return pairs(self.libs) end setmetatable(LibStub, { __call = LibStub.GetLibrary })
end

local LBFGroup
local LBF = LibStub('LibButtonFacade', true)

local bBars_TabBinder = function()
	local eventf = CreateFrame("frame")
	eventf:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	eventf:RegisterEvent("PLAYER_REGEN_ENABLED")
	eventf:RegisterEvent("DUEL_REQUESTED")
	eventf:RegisterEvent("DUEL_FINISHED")
	eventf:RegisterEvent("CHAT_MSG_SYSTEM")
	eventf.isFail = false

	eventf:SetScript("OnEvent", function(self, event, ...)
		local arg1, arg2, arg3 = ...

		if event=="CHAT_MSG_SYSTEM" then
			if arg1 == ERR_DUEL_REQUESTED then event = "DUEL_REQUESTED" end
		end
		if event=="ZONE_CHANGED_NEW_AREA" or (event=="PLAYER_REGEN_ENABLED" and self.isFail) or event=="DUEL_REQUESTED" or event=="DUEL_FINISHED" then
			local which, pvpType = GetCurrentBindingSet(), GetZonePVPInfo() or ""
			local _, zoneType = IsInInstance()
			if zoneType==nil then zoneType="" end
			
			tabKey = GetBindingKey("TARGETNEARESTENEMYPLAYER")
			if tabKey == nil then tabKey = GetBindingKey("TARGETNEARESTENEMY") end
			if tabKey == nil then tabKey = "TAB" end
			lastTabKey = GetBindingKey("TARGETPREVIOUSENEMYPLAYER")
			if lastTabKey == nil then lastTabKey = GetBindingKey("TARGETPREVIOUSENEMY") end
			if lastTabKey == nil then lastTabKey = "SHIFT-TAB" end

			local CurrentBind = GetBindingAction(tabKey)

			if (pvpType:find("arena") or pvpType:find("combat")) or zoneType:find("pvp") or event=="DUEL_REQUESTED" then
				if CurrentBind ~= "TARGETNEARESTENEMYPLAYER" then
					local success = SetBinding(tabKey, "TARGETNEARESTENEMYPLAYER")
					SetBinding(lastTabKey, "TARGETPREVIOUSENEMYPLAYER")
					if success == 1 then
						if self.isFail then
							print(""..tabKey.." set to |cff00ccffenemy players only|r.")
						end
						SaveBindings(which)
						self.isFail = false
					else
						print(""..tabKey.." binding |cffff0000failed|r to update.")
						self.isFail = true
					end
				end
			else
				if CurrentBind ~= "TARGETNEARESTENEMY" then
					local success = SetBinding(tabKey, "TARGETNEARESTENEMY")
					SetBinding(lastTabKey, "TARGETPREVIOUSENEMY")
					if success == 1 then
						if self.isFail then
							print(""..tabKey.." set to |cff00ccffall enemies|r.")
						end
						SaveBindings(which)
						self.isFail = false
					else
						print(""..tabKey.." binding |cffff0000failed|r to update.")
						self.isFail = true
					end
				end
			end
		end
	end)	
end

local bBars_OmniCC = function()
	--[[ An edited lightweight OmniCC for Tukui
		A featureless, 'pure' version of OmniCC.
		This version should work on absolutely everything, but I've removed pretty much all of the options
	--]]

	if IsAddOnLoaded("OmniCC") or IsAddOnLoaded("ncCooldown") then return end

	--constants!
	local ICON_SIZE = 36 --the normal size for an icon (don't change this)
	local DAY, HOUR, MINUTE = 86400, 3600, 60 --used for formatting text
	local DAYISH, HOURISH, MINUTEISH = 3600 * 23.5, 60 * 59.5, 59.5 --used for formatting text at transition points
	local HALFDAYISH, HALFHOURISH, HALFMINUTEISH = DAY/2 + 0.5, HOUR/2 + 0.5, MINUTE/2 + 0.5 --used for calculating next update times

	--configuration settings
	local FONT_SIZE = 20 --the base font size to use at a scale of 1
	local MIN_SCALE = 0.5 --the minimum scale we want to show cooldown counts at, anything below this will be hidden
	local MIN_DURATION = 2.5 --the minimum duration to show cooldown text for
	local EXPIRING_DURATION = .3 --the minimum number of seconds a cooldown must be to use to display in the expiring format

	local EXPIRING_FORMAT = '|cffff0000%.1f|r' --format for timers that are soon to expire
	local SECONDS_FORMAT = '|cffffff00%d|r' --format for timers that have seconds remaining
	local MINUTES_FORMAT = '|cffffffff%dm|r' --format for timers that have minutes remaining
	local HOURS_FORMAT = '|cffaaffff%dh|r' --format for timers that have hours remaining
	local DAYS_FORMAT = '|cffaaaaff%dh|r' --format for timers that have days remaining

	--local bindings!
	local floor = math.floor
	local min = math.min
	local GetTime = GetTime

	--returns both what text to display, and how long until the next update
	local function getTimeText(s)
		--format text as seconds when below a minute
		if s < MINUTEISH then
			local seconds = tonumber(round(s))
			if seconds > EXPIRING_DURATION then
				return SECONDS_FORMAT, seconds, s - (seconds - 0.51)
			else
				return EXPIRING_FORMAT, s, 0.051
			end
		--format text as minutes when below an hour
		elseif s < HOURISH then
			local minutes = tonumber(round(s/MINUTE))
			return MINUTES_FORMAT, minutes, minutes > 1 and (s - (minutes*MINUTE - HALFMINUTEISH)) or (s - MINUTEISH)
		--format text as hours when below a day
		elseif s < DAYISH then
			local hours = tonumber(round(s/HOUR))
			return HOURS_FORMAT, hours, hours > 1 and (s - (hours*HOUR - HALFHOURISH)) or (s - HOURISH)
		--format text as days
		else
			local days = tonumber(round(s/DAY))
			return DAYS_FORMAT, days,  days > 1 and (s - (days*DAY - HALFDAYISH)) or (s - DAYISH)
		end
	end

	--stops the timer
	local function Timer_Stop(self)
		self.enabled = nil
		self:Hide()
	end

	--forces the given timer to update on the next frame
	local function Timer_ForceUpdate(self)
		self.nextUpdate = 0
		self:Show()
	end

	--adjust font size whenever the timer's parent size changes
	--hide if it gets too tiny
	local function Timer_OnSizeChanged(self, width, height)
		local fontScale = round(width) / ICON_SIZE
		if fontScale == self.fontScale then
			return
		end

		self.fontScale = fontScale
		if fontScale < MIN_SCALE then
			self:Hide()
		else
			self.text:SetFont(select(1, ChatFontNormal:GetFont()), fontScale * FONT_SIZE, 'OUTLINE')
			if self.enabled then
				Timer_ForceUpdate(self)
			end
		end
	end

	--update timer text, if it needs to be
	--hide the timer if done
	local function Timer_OnUpdate(self, elapsed)
		if self.nextUpdate > 0 then
			self.nextUpdate = self.nextUpdate - elapsed
		else
			local remain = self.duration - (GetTime() - self.start)
			if tonumber(round(remain)) > 0 then
				if (self.fontScale * self:GetEffectiveScale() / UIParent:GetScale()) < MIN_SCALE then
					self.text:SetText('')
					self.nextUpdate  = 1
				else
					local formatStr, time, nextUpdate = getTimeText(remain)
					self.text:SetFormattedText(formatStr, time)
					self.nextUpdate = nextUpdate
				end
			else
				Timer_Stop(self)
			end
		end
	end

	--returns a new timer object
	local function Timer_Create(self)
		--a frame to watch for OnSizeChanged events
		--needed since OnSizeChanged has funny triggering if the frame with the handler is not shown
		local scaler = CreateFrame('Frame', nil, self)
		scaler:SetAllPoints(self)

		local timer = CreateFrame('Frame', nil, scaler); timer:Hide()
		timer:SetAllPoints(scaler)
		timer:SetScript('OnUpdate', Timer_OnUpdate)

		local text = timer:CreateFontString(nil, 'OVERLAY')
		text:SetPoint("CENTER", 2, 0)
		text:SetJustifyH("CENTER")
		timer.text = text

		Timer_OnSizeChanged(timer, scaler:GetSize())
		scaler:SetScript('OnSizeChanged', function(self, ...) Timer_OnSizeChanged(timer, ...) end)

		self.timer = timer
		return timer
	end

	--hook the SetCooldown method of all cooldown frames
	--ActionButton1Cooldown is used here since its likely to always exist
	--and I'd rather not create my own cooldown frame to preserve a tiny bit of memory
	local function Timer_Start(self, start, duration)
		if self.noOCC then return end
		--start timer
		if start > 0 and duration > MIN_DURATION then
			local timer = self.timer or Timer_Create(self)
			timer.start = start
			timer.duration = duration
			timer.enabled = true
			timer.nextUpdate = 0
			if timer.fontScale >= MIN_SCALE then timer:Show() end
		--stop timer
		else
			local timer = self.timer
			if timer then
				Timer_Stop(timer)
			end
		end
	end

	local active = {}
	local hooked = {}

	local function cooldown_OnShow(self)
		active[self] = true
	end

	local function cooldown_OnHide(self)
		active[self] = nil
	end

	local function cooldown_ShouldUpdateTimer(self, start, duration)
		local timer = self.timer
		if not timer then
			return true
		end
		return timer.start ~= start
	end

	local function cooldown_Update(self)
		local button = self:GetParent()
		local start, duration, enable = GetActionCooldown(button.action)

		if cooldown_ShouldUpdateTimer(self, start, duration) then
			Timer_Start(self, start, duration)
		end
	end

	local EventWatcher = CreateFrame("Frame")
	EventWatcher:Hide()
	EventWatcher:SetScript("OnEvent", function(self, event)
		for cooldown in pairs(active) do
			cooldown_Update(cooldown)
		end
	end)
	EventWatcher:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")

	local function actionButton_Register(frame)
		local cooldown = frame.cooldown
		if not hooked[cooldown] then
			cooldown:HookScript("OnShow", cooldown_OnShow)
			cooldown:HookScript("OnHide", cooldown_OnHide)
			hooked[cooldown] = true
		end
	end

	if _G["ActionBarButtonEventsFrame"].frames then
		for i, frame in pairs(_G["ActionBarButtonEventsFrame"].frames) do
			actionButton_Register(frame)
		end
	end
	
	for i, button in pairs(ActionBarButtonEventsFrame.frames) do
		actionButton_Register(button)
	end
	

	hooksecurefunc(getmetatable(ActionButton1Cooldown).__index, "SetCooldown", Timer_Start)
	hooksecurefunc("ActionBarButtonEventsFrame_RegisterFrame", actionButton_Register)
	hooksecurefunc('SetActionUIButton', actionButton_Register)
end

local bBars_HoverBind = function()
	ncHoverBind = _G.ncHoverBind or CreateFrame('Frame', 'ncHoverBind', bBars)
	ncHoverBind.localmacros = 0
	if InCombatLockdown() then print('You can\'t bind keys in combat.') return end
	if not ncHoverBind.loaded then
		local find = string.find
		local _G = getfenv(0)

		ncHoverBind:SetFrameStrata('DIALOG')
		ncHoverBind:EnableMouse(true)
		ncHoverBind:EnableKeyboard(true)
		ncHoverBind:EnableMouseWheel(true)
		ncHoverBind:Hide()

		local elapsed = 0
		GameTooltip:HookScript('OnUpdate', function(self, e)
			elapsed = elapsed + e
			if elapsed < .2 then return else elapsed = 0 end
			if (not self.comparing and IsModifiedClick('COMPAREITEMS')) then
				GameTooltip_ShowCompareItem(self)
				self.comparing = true
			elseif ( self.comparing and not IsModifiedClick('COMPAREITEMS')) then
				for _, frame in pairs(self.shoppingTooltips) do
					frame:Hide()
				end
				self.comparing = false
			end
		end)
		hooksecurefunc(GameTooltip, 'Hide', function(self) for _, tt in pairs(self.shoppingTooltips) do tt:Hide() end end)

		ncHoverBind:SetScript('OnEvent', function(self) self:Deactivate(false) end)
		ncHoverBind:SetScript('OnLeave', function(self) self:HideFrame() end)
		ncHoverBind:SetScript('OnKeyUp', function(self, key) self:Listener(key) end)
		ncHoverBind:SetScript('OnMouseUp', function(self, key) self:Listener(key) end)
		ncHoverBind:SetScript('OnMouseWheel', function(self, delta) if delta>0 then self:Listener('MOUSEWHEELUP') else self:Listener('MOUSEWHEELDOWN') end end)

		function ncHoverBind:Update(b, spellmacro)
			if not self.enabled or InCombatLockdown() then return end
			self.button = b
			self.spellmacro = spellmacro
			
			self:ClearAllPoints()
			self:SetAllPoints(b)
			self:Show()
			
			ShoppingTooltip1:Hide()
			
			if spellmacro=='SPELL' then
				self.button.id = SpellBook_GetSpellID(self.button:GetID())
				self.button.name = GetSpellName(self.button.id, SpellBookFrame.bookType)
				
				GameTooltip:AddLine('Trigger')
				GameTooltip:Show()
				GameTooltip:SetScript('OnHide', function(self)
					self:SetOwner(ncHoverBind, 'ANCHOR_NONE')
					self:SetPoint('BOTTOM', ncHoverBind, 'TOP', 0, 1)
					self:AddLine(ncHoverBind.button.name, 1, 1, 1)
					ncHoverBind.button.bindings = {GetBindingKey(spellmacro..' '..ncHoverBind.button.name)}
					if #ncHoverBind.button.bindings == 0 then
						self:AddLine('No bindings set.', .6, .6, .6)
					else
						self:AddDoubleLine('Binding', 'Key', .6, .6, .6, .6, .6, .6)
						for i = 1, #ncHoverBind.button.bindings do
							self:AddDoubleLine(i, ncHoverBind.button.bindings[i])
						end
					end
					self:Show()
					self:SetScript('OnHide', nil)
				end)
			elseif spellmacro=='MACRO' then
				self.button.id = self.button:GetID()
				
				if ncHoverBind.localmacros==1 then self.button.id = self.button.id + 36 end
				
				self.button.name = GetMacroInfo(self.button.id)
				
				GameTooltip:SetOwner(ncHoverBind, 'ANCHOR_NONE')
				GameTooltip:SetPoint('BOTTOM', ncHoverBind, 'TOP', 0, 1)
				GameTooltip:AddLine(ncHoverBind.button.name, 1, 1, 1)
				
				ncHoverBind.button.bindings = {GetBindingKey(spellmacro..' '..ncHoverBind.button.name)}
					if #ncHoverBind.button.bindings == 0 then
						GameTooltip:AddLine('No bindings set.', .6, .6, .6)
					else
						GameTooltip:AddDoubleLine('Binding', 'Key', .6, .6, .6, .6, .6, .6)
						for i = 1, #ncHoverBind.button.bindings do
							GameTooltip:AddDoubleLine('Binding'..i, ncHoverBind.button.bindings[i], 1, 1, 1)
						end
					end
				GameTooltip:Show()
			elseif spellmacro=='STANCE' or spellmacro=='PET' then
				self.button.id = tonumber(b:GetID())
				self.button.name = b:GetName()
				
				if not self.button.name then return end
				
				if not self.button.id or self.button.id < 1 or self.button.id > (spellmacro=='STANCE' and 10 or 12) then
					self.button.bindstring = 'CLICK '..self.button.name..':LeftButton'
				else
					self.button.bindstring = (spellmacro=='STANCE' and 'SHAPESHIFTBUTTON' or 'BONUSACTIONBUTTON')..self.button.id
				end
				
				GameTooltip:AddLine('Trigger')
				GameTooltip:Show()
				GameTooltip:SetScript('OnHide', function(self)
					self:SetOwner(ncHoverBind, 'ANCHOR_NONE')
					self:SetPoint('BOTTOM', ncHoverBind, 'TOP', 0, 1)
					self:AddLine(ncHoverBind.button.name, 1, 1, 1)
					ncHoverBind.button.bindings = {GetBindingKey(ncHoverBind.button.bindstring)}
					if #ncHoverBind.button.bindings == 0 then
						self:AddLine('No bindings set.', .6, .6, .6)
					else
						self:AddDoubleLine('Binding', 'Key', .6, .6, .6, .6, .6, .6)
						for i = 1, #ncHoverBind.button.bindings do
							self:AddDoubleLine(i, ncHoverBind.button.bindings[i])
						end
					end
					self:Show()
					self:SetScript('OnHide', nil)
				end)
			else
				self.button.action = tonumber(b.action)
				self.button.name = b:GetName()
				
				if not self.button.name then return end
				
				if not self.button.action or self.button.action < 1 or self.button.action > 132 then
					self.button.bindstring = 'CLICK '..self.button.name..':LeftButton'
				else
					local modact = 1+(self.button.action-1)%12
					if self.button.action < 25 or self.button.action > 72 then
						self.button.bindstring = 'ACTIONBUTTON'..modact
					elseif self.button.action < 73 and self.button.action > 60 then
						self.button.bindstring = 'MULTIACTIONBAR1BUTTON'..modact
					elseif self.button.action < 61 and self.button.action > 48 then
						self.button.bindstring = 'MULTIACTIONBAR2BUTTON'..modact
					elseif self.button.action < 49 and self.button.action > 36 then
						self.button.bindstring = 'MULTIACTIONBAR4BUTTON'..modact
					elseif self.button.action < 37 and self.button.action > 24 then
						self.button.bindstring = 'MULTIACTIONBAR3BUTTON'..modact
					end
				end
				
				GameTooltip:AddLine('Trigger')
				GameTooltip:Show()
				GameTooltip:SetScript('OnHide', function(self)
					self:SetOwner(ncHoverBind, 'ANCHOR_NONE')
					self:SetPoint('BOTTOM', ncHoverBind, 'TOP', 0, 1)
					self:AddLine(ncHoverBind.button.name, 1, 1, 1)
					ncHoverBind.button.bindings = {GetBindingKey(ncHoverBind.button.bindstring)}
					if #ncHoverBind.button.bindings == 0 then
						self:AddLine('No bindings set.', .6, .6, .6)
					else
						self:AddDoubleLine('Binding', 'Key', .6, .6, .6, .6, .6, .6)
						for i = 1, #ncHoverBind.button.bindings do
							self:AddDoubleLine(i, ncHoverBind.button.bindings[i])
						end
					end
					self:Show()
					self:SetScript('OnHide', nil)
				end)
			end
		end

		function ncHoverBind:Listener(key)
			if key == 'ESCAPE' or key == 'RightButton' then
				for i = 1, #self.button.bindings do
					SetBinding(self.button.bindings[i])
				end
				print('All keybindings cleared for |cff00ff00'..self.button.name..'|r.')
				self:Update(self.button, self.spellmacro)
				if self.spellmacro~='MACRO' then GameTooltip:Hide() end
				return
			end
			
			if key == 'LSHIFT'
			or key == 'RSHIFT'
			or key == 'LCTRL'
			or key == 'RCTRL'
			or key == 'LALT'
			or key == 'RALT'
			or key == 'UNKNOWN'
			or key == 'LeftButton'
			or key == 'MiddleButton'
			then return end
			
			if key == 'Button4' then key = 'BUTTON4' end
			if key == 'Button5' then key = 'BUTTON5' end
			
			local alt = IsAltKeyDown() and 'ALT-' or ''
			local ctrl = IsControlKeyDown() and 'CTRL-' or ''
			local shift = IsShiftKeyDown() and 'SHIFT-' or ''
			
			if not self.spellmacro or self.spellmacro=='PET' or self.spellmacro=='STANCE' then
				SetBinding(alt..ctrl..shift..key, self.button.bindstring)
			else
				SetBinding(alt..ctrl..shift..key, self.spellmacro..' '..self.button.name)
			end
			print(alt..ctrl..shift..key..' |cff00ff00bound to |r'..self.button.name..'.')
			self:Update(self.button, self.spellmacro)
			if self.spellmacro~='MACRO' then GameTooltip:Hide() end
		end
		
		function ncHoverBind:HideFrame()
			self:ClearAllPoints()
			self:Hide()
			GameTooltip:Hide()
		end
		
		function ncHoverBind:Activate()
			self.enabled = true
			self:RegisterEvent('PLAYER_REGEN_DISABLED')
		end
		
		function ncHoverBind:Deactivate(save)
			if save then
				SaveBindings(2)
				print('All keybindings have been saved.')
			else
				LoadBindings(2)
				print('All newly set keybindings have been discarded.')
			end
			self.enabled = false
			self:HideFrame()
			self:UnregisterEvent('PLAYER_REGEN_DISABLED')
			StaticPopup_Hide('KEYBIND_MODE')
		end

		StaticPopupDialogs['KEYBIND_MODE'] = {
			preferredIndex = STATICPOPUP_NUMDIALOGS,
			text = 'Hover your mouse over any actionbutton to bind it. Press the escape key or right click to clear the current actionbutton\'s keybinding.',
			button1 = 'Save bindings',
			button2 = 'Discard bindings',
			OnAccept = function() ncHoverBind:Deactivate(true) end,
			OnCancel = function() ncHoverBind:Deactivate(false) end,
			timeout = 0,
			whileDead = 1,
			hideOnEscape = false
		}

		-- REGISTERING
		local stance = _G['ShapeshiftButton1'] and _G['ShapeshiftButton1']:GetScript('OnClick') or _G['Button1'] and _G['Button1']:GetScript('OnClick')
		local pet = PetActionButton1:GetScript('OnClick')
		local button = SecureActionButton_OnClick

		local function register(val)
			if val.IsProtected and val.GetObjectType and val.GetScript and val:GetObjectType()=='CheckButton' and val:IsProtected() then
				local script = val:GetScript('OnClick')
				if script==button then
					val:HookScript('OnEnter', function(self) ncHoverBind:Update(self) end)
				elseif script==stance then
					val:HookScript('OnEnter', function(self) ncHoverBind:Update(self, 'STANCE') end)
				elseif script==pet then
					val:HookScript('OnEnter', function(self) ncHoverBind:Update(self, 'PET') end)
				end
			end
		end

		local val = EnumerateFrames()
		while val do
			register(val)
			val = EnumerateFrames(val)
		end

		for i=1,12 do
			local b = _G['SpellButton'..i]
			b:HookScript('OnEnter', function(self) ncHoverBind:Update(self, 'SPELL') end)
		end
		
		local function registermacro()
			for i=1,36 do
				local b = _G['MacroButton'..i]
				b:HookScript('OnEnter', function(self) ncHoverBind:Update(self, 'MACRO') end)
			end
			MacroFrameTab1:HookScript('OnMouseUp', function() ncHoverBind.localmacros = 0 end)
			MacroFrameTab2:HookScript('OnMouseUp', function() ncHoverBind.localmacros = 1 end)
		end
		
		if not IsAddOnLoaded('Blizzard_MacroUI') then
			hooksecurefunc('LoadAddOn', function(addon)
				if addon=='Blizzard_MacroUI' then
					registermacro()
				end
			end)
		else
			registermacro()
		end
		ncHoverBind.loaded = 1
	end
	if not ncHoverBind.enabled then
		ncHoverBind:Activate()
		StaticPopup_Show('KEYBIND_MODE')
	end
end

local bBars_Stick = function(frame, otherFrame, tolerance, xOff, yOff)
	local function FrameIsDependentOnFrame(frame, otherFrame)
		if (frame and otherFrame) then
			if frame == otherFrame then return true end
			local points = frame:GetNumPoints()
			for i = 1, points do
				local parent = select(2, frame:GetPoint(i))
				if FrameIsDependentOnFrame(parent, otherFrame) then
					return true
				end
			end
		end
	end

	local function CanAttach(frame, otherFrame)
		if not(frame and otherFrame) then
			return
		elseif frame:GetNumPoints() == 0 or otherFrame:GetNumPoints() == 0 then
			return
		elseif frame:GetWidth() == 0 or frame:GetHeight() == 0 or otherFrame:GetWidth() == 0 or otherFrame:GetHeight() == 0 then
			return
		elseif FrameIsDependentOnFrame(otherFrame, frame) then
			return
		end
		return true
	end

	local function AttachToTop(frame, otherFrame, distLeft, distRight, distCenter, offset)
		frame:ClearAllPoints()
		if distLeft < distCenter and distLeft < distRight then
			frame:SetPoint('BOTTOMLEFT', otherFrame, 'TOPLEFT', 0, Fix(offset))
		elseif distRight < distCenter and distRight < distLeft then
			frame:SetPoint('BOTTOMRIGHT', otherFrame, 'TOPRIGHT', 0, Fix(offset))
		else
			frame:SetPoint('BOTTOM', otherFrame, 'TOP', 0, Fix(offset))
		end
	end

	local function AttachToBottom(frame, otherFrame, distLeft, distRight, distCenter, offset)
		frame:ClearAllPoints()
		if distLeft < distCenter and distLeft < distRight then
			frame:SetPoint('TOPLEFT', otherFrame, 'BOTTOMLEFT', 0, -Fix(offset))
		elseif distRight < distCenter and distRight < distLeft then
			frame:SetPoint('TOPRIGHT', otherFrame, 'BOTTOMRIGHT', 0, -Fix(offset))
		else
			frame:SetPoint('TOP', otherFrame, 'BOTTOM', 0, -Fix(offset))
		end
	end

	local function AttachToLeft(frame, otherFrame, distTop, distBottom, distCenter, offset)
		frame:ClearAllPoints()
		if distBottom < distTop and distBottom < distCenter then
			frame:SetPoint('BOTTOMRIGHT', otherFrame, 'BOTTOMLEFT', -Fix(offset), 0)
		elseif distTop < distBottom and distTop < distCenter then
			frame:SetPoint('TOPRIGHT', otherFrame, 'TOPLEFT', -Fix(offset), 0)
		else
			frame:SetPoint('RIGHT', otherFrame, 'LEFT', -Fix(offset), 0)
		end
	end

	local function AttachToRight(frame, otherFrame, distTop, distBottom, distCenter, offset)
		frame:ClearAllPoints()
		if distBottom < distTop and distBottom < distCenter then
			frame:SetPoint('BOTTOMLEFT', otherFrame, 'BOTTOMRIGHT', Fix(offset), 0)
		elseif distTop < distBottom and distTop < distCenter then
			frame:SetPoint('TOPLEFT', otherFrame, 'TOPRIGHT', Fix(offset), 0)
		else
			frame:SetPoint('LEFT', otherFrame, 'RIGHT', Fix(offset), 0)
		end
	end

	function StickFrames(frame, otherFrame, tolerance, xOff, yOff)
		local xOff, yOff = xOff or 0, yOff or 0
		local left = frame:GetLeft()
		local right = frame:GetRight()
		local top = frame:GetTop()
		local bottom = frame:GetBottom()
		local centerX, centerY = frame:GetCenter()

		if not CanAttach(frame, otherFrame) then return end
		
		if left and right and top and bottom and centerX then
			local oScale = otherFrame:GetScale()
			left = left / oScale
			right = right / oScale
			top = top / oScale
			bottom = bottom /oScale
			centerX = centerX / oScale
			centerY = centerY / oScale
		else return end

		local oLeft = otherFrame:GetLeft()
		local oRight = otherFrame:GetRight()
		local oTop = otherFrame:GetTop()
		local oBottom = otherFrame:GetBottom()
		local oCenterX, oCenterY = otherFrame:GetCenter()

		if oLeft and oRight and oTop and oBottom and oCenterX then
			local scale = frame:GetScale()
			oCenterX = oCenterX / scale
			oCenterY = oCenterY / scale
			oLeft = oLeft / scale
			oRight = oRight / scale
			oTop = oTop / scale
			oBottom = oBottom / scale
		else return end

		if(oLeft - tolerance <= left and oRight + tolerance >= right) or (left - tolerance <= oLeft and right + tolerance >= oRight)then
			local distCenter = math.abs(oCenterX - centerX)
			local distLeft = math.abs(oLeft - left)
			local distRight = math.abs(right - oRight)

			if math.abs(oTop - bottom) <= tolerance then 
				return AttachToTop(frame, otherFrame, distLeft, distRight, distCenter, yOff) 
			elseif math.abs(oBottom - top) <= tolerance then 
				return AttachToBottom(frame, otherFrame, distLeft, distRight, distCenter, yOff)
			end
		end

		if(oTop + tolerance >= top and oBottom - tolerance <= bottom) or (top + tolerance >= oTop and bottom - tolerance <= oBottom) then
			local distCenter = math.abs(oCenterY - centerY)
			local distTop = math.abs(oTop - top)
			local distBottom = math.abs(oBottom - bottom)
			
			if math.abs(oLeft - right) <= tolerance then return AttachToLeft(frame, otherFrame, distTop, distBottom, distCenter, xOff) end
			if math.abs(oRight - left) <= tolerance then return AttachToRight(frame, otherFrame, distTop, distBottom, distCenter, xOff) end
		end
	end
	StickFrames(frame, otherFrame, tolerance, xOff, yOff)
end

-- hide blizzard frames region
local bBars_HideBlizzard = function()
	local hider = _G['BlizzardHider'] or CreateFrame('Frame', 'BlizzardHider', bBars)
	hider:Hide()
	
	for i,v in pairs({'OverrideActionBar','ActionBarController', 'MainMenuBar', 'MultiBarBottomLeft', 'MultiBarBottomRight', 'MultiBarLeft', 'MultiBarRight', 'ShapeshiftBarFrame', 'BonusActionBarFrame', 'PossessBarFrame', 'MainMenuExpBar', 'MainMenuBarArtFrame'})do
		if _G[v] then
			_G[v]:UnregisterAllEvents()
			_G[v]:SetScript('OnUpdate', nil)
			_G[v]:SetParent(hider)
			_G[v]:Hide()
		end
	end
	
	if PetBattleFrame then
		PetBattleFrame.BottomFrame:SetParent(hider)
		PetBattleFrame.BottomFrame.PetSelectionFrame:SetParent(bBars)
	end
	
	for i,v in pairs({'MultiBarRight', 'MultiBarLeft', 'MultiBarBottomLeft', 'MultiBarBottomRight', 'MainMenuBar', 'ShapeshiftBarFrame', 'PossessBarFrame', 'MultiCastActionBarFrame', 'PETACTIONBAR_YPOS', 'MULTICASTACTIONBAR_YPOS'}) do
		if UIPARENT_MANAGED_FRAME_POSITIONS[v] then
			UIPARENT_MANAGED_FRAME_POSITIONS[v] = nil
		end
	end

	MainMenuBarArtFrame:RegisterEvent('BAG_UPDATE')
	MainMenuBarArtFrame:RegisterEvent('CURRENCY_DISPLAY_UPDATE')
	MainMenuBarArtFrame:RegisterEvent('UNIT_LEVEL')
	
	if not _G['VehicleSeatIndicator']:IsUserPlaced() then _G['VehicleSeatIndicator']:SetPoint('TOPRIGHT', MinimapCluster, 'BOTTOMRIGHT', 0, -13) end
	for i = 1, 6 do 
		if _G['VehicleMenuBarActionButton'..i] then _G['VehicleMenuBarActionButton'..i]:UnregisterAllEvents() end 
	end
	for i = 1, 12 do 
		if _G['ActionButton'..i] then _G['ActionButton'..i]:UnregisterAllEvents() _G['ActionButton'..i]:SetScript('OnUpdate', nil) end
		if _G['BonusActionButton'..i] then _G['BonusActionButton'..i]:UnregisterAllEvents() _G['BonusActionButton'..i]:SetScript('OnUpdate', nil) end
		if _G['MultiCastActionButton'..i] then _G['MultiCastActionButton'..i]:UnregisterEvent('UPDATE_BINDINGS') _G['MultiCastActionButton'..i]:SetScript('OnUpdate', nil) end
	end
	
	if PlayerTalentFrame then
		PlayerTalentFrame:UnregisterEvent('ACTIVE_TALENT_GROUP_CHANGED')
	else
		hooksecurefunc('TalentFrame_LoadUI', function() PlayerTalentFrame:UnregisterEvent('ACTIVE_TALENT_GROUP_CHANGED') end)
	end

	MultiActionBar_UpdateGrid = nullfunc
end
-- end region

local bBars_createTitle = function(self)
	self.title = self.title or self:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
	self.title:SetPoint('CENTER')
	self.title:SetText(string.sub(self:GetName(), 2))
	self.title:SetDrawLayer('OVERLAY', 7)
	self.title:Hide()
end

local bBars_UpdateBind = function(self, buttonType)
	local bname = self:GetName()
	local macro = _G[bname..'Name'] or self:CreateFontString(bname..'Name', 'ARTWORK')
	local hotkey = _G[bname..'HotKey'] or self:CreateFontString(bname..'HotKey', 'ARTWORK')
	macro:SetFont(bBars_font, fontSize, fontStyle)
	macro:SetShadowOffset(0, 0)
	macro:SetText(macro:GetText() and macro:GetText():lower() or '')
	
	if bBarsDB[pname].keybindings == 1 then
		local bind = self:GetAttribute('binding')
		local key
		
		if bind then
			key = GetBindingKey(bind) 
		else 
			key = GetBindingKey('CLICK '..bname..':LeftButton')
		end
		
		if key then
			key = lower(key)
			key = replace(key, '^button(%d+)$', 'b%1')
			key = replace(key, 'shift%-', 's')
			key = replace(key, 'alt%-', 'a')
			key = replace(key, 'ctrl%-', 'c')
			key = replace(key, 'numpad%-', 'n')
			
			hotkey:SetText((hotkey:GetText() == RANGE_INDICATOR) and '' or GetBindingText(key, 'KEY_', 1) and GetBindingText(key, 'KEY_', 1) or '')
			hotkey:Show()
		else
			hotkey:Hide()
		end
	else
		hotkey:Hide()
	end
	
	if bBarsDB[pname].macroses == 1 then macro:Show() else macro:Hide() end
end

hooksecurefunc('ActionButton_UpdateHotkeys', bBars_UpdateBind)

local bBars_setStyle = function(self)
	--if true then return end
	if self.actionbutton then 
		if self.action and select(1, GetActionInfo(self.action)) then 
			self:SetAlpha(1) 
		else 
			self:SetAlpha(.2)
		end
	end
	
	if InCombatLockdown() then return end
	
	if self.styled then return end
	
	do
		local button 	= self
		local bname 	= button:GetName()
		local icon 		= _G[bname..'Icon']
		local border 	= _G[bname..'Border']
		local fbg 		= _G[bname..'FloatingBG']
		local nt 		= _G[bname..'NormalTexture'] and _G[bname..'NormalTexture'] or button:GetNormalTexture() and button:GetNormalTexture() or nil
		local flash 	= _G[bname..'Flash']
		local count 	= _G[bname..'Count']

		button:SetScale(UIParent:GetScale())
		
		local macro = _G[bname..'Name'] or self:CreateFontString(bname..'Name', 'ARTWORK')
		local hotkey = _G[bname..'HotKey'] or self:CreateFontString(bname..'HotKey', 'ARTWORK')
		
		if count then
			count:SetFont(bBars_font, fontSize, fontStyle)
			count:ClearAllPoints()
			count:SetPoint('BOTTOMRIGHT', 0, 2)
			count.SetTextHeight = nullfunc
			count.SetTextWidth = nullfunc
			count.SetFont = nullfunc
		end
		
		if hotkey then
			hotkey:SetFont(bBars_font, fontSize, fontStyle)
			hotkey:ClearAllPoints()
			hotkey:SetPoint('TOPRIGHT', 0, -1)
			hotkey:SetJustifyH('RIGHT')
			hotkey.ClearAllPoints = nullfunc
			hotkey.SetPoint = nullfunc
			hotkey.SetTextHeight = nullfunc
			hotkey.SetTextWidth = nullfunc
			hotkey.SetFont = nullfunc
			hotkey:SetDrawLayer('OVERLAY', 9)
			
			macro:SetFont(bBars_font, fontSize, fontStyle)
			macro:SetSize(button:GetWidth()-2, 10)
			macro.SetFont = nullfunc
		end
		
		if fbg then fbg:Hide() fbg.Show = nullfunc end
		local pushed, checked, highlight = button.GetPushedTexture and button:GetPushedTexture() or nil, button.GetCheckedTexture and button:GetCheckedTexture() or nil, button.GetHighlightTexture and button:GetHighlightTexture() or nil
		
		if LBF and LBFGroup and nt and not lower(bname):find('micro') then return end
		
		if icon then 
			icon:SetTexCoord(.12, .88, .12, .88) 
			icon:ClearAllPoints()
			icon:SetPoint('TOPLEFT', 1, -1)
			icon:SetPoint('BOTTOMRIGHT', -1, 1)
			icon:SetDrawLayer('BACKGROUND', 1)
		end
		
		if button.SelectedHighlight then
			button.SelectedHighlight:SetTexture(0.9,0.8,0.1, .2)
			button.SelectedHighlight:ClearAllPoints()
			button.SelectedHighlight:SetAllPoints()
			button.SelectedHighlight:SetDrawLayer('BACKGROUND', 0)
		end

		if button.icon then
			button.icon:SetTexCoord(.08, .95, .08, .95)
			button.icon:SetDrawLayer('BACKGROUND', 0)
		end
		
		if button.AdditionalIcon then
			button.AdditionalIcon:SetTexCoord(.1, .9, .1, .9)
			button.AdditionalIcon:SetDrawLayer('BACKGROUND', 0)
		end
		
		if button.Icon then
			button.Icon:SetTexCoord(.1, .9, .1, .9)
			button.Icon:SetDrawLayer('BACKGROUND', 1)
		end

		if border then 
			border:SetTexture(0.9,0.8,0.1,1) 
			border:SetAllPoints()
		else
			if bname:find('MicroButton') then
				border = _G[bname..'Border'] or button:CreateTexture(bname..'Border', 'BACKGROUND')
				border:SetTexture(0.9,0.8,0.1,1)
				border:SetAllPoints()
			end
		end
		
		if flash then
			flash:SetTexture(0.9,0.8,0.1,0.3)
			flash:ClearAllPoints()
			flash:SetAllPoints(button)
			flash.SetPoint = nullfunc
			flash.SetTexture = nullfunc
		end
		
		if pushed then
			pushed:SetTexture(0.9,0.8,0.1,0.3)
			pushed:SetPoint('TOPLEFT', button, 1, -1)
			pushed:SetPoint('BOTTOMRIGHT', button, -1, 1)
			button:SetPushedTexture(pushed)
			button.SetPushedTexture = nullfunc
		end
		
		if checked then
			checked:SetTexture(1, .3, 0, .7)
			checked:SetPoint('TOPLEFT', button, 1, -1)
			checked:SetPoint('BOTTOMRIGHT', button, -1, 1)
			button:SetCheckedTexture(checked)
			button.SetCheckedTexture = nullfunc
		end

		if highlight then
			highlight:SetTexture(1,1,1,0.3)
			highlight:SetPoint('TOPLEFT', button, 1, -1)
			highlight:SetPoint('BOTTOMRIGHT', button, -1, 1)
			button:SetHighlightTexture(highlight)
			button.SetHighlightTexture = nullfunc
		end
		
		if nt then
			if not lower(bname):find('microbutton') and not lower(bname):find('vehicle') then
				nt:SetTexture(nil)
				nt:SetTexCoord(.5, .5, .5, .5)
				nt.SetTexture = nullfunc
				if button.nt == nil then
					button.nt = button:CreateTexture(nil, 'OVERLAY')
					button.nt:SetAllPoints()
					button.nt:SetTexture('Interface\\Buttons\\WHITE8x8')
					button.nt:SetVertexColor(0,0,0)
					button.nt:SetDrawLayer('BACKGROUND', -7)
				end
			else
				if not lower(bname):find('vehicle') then
					nt:SetTexCoord(.15, .84, .48, .9)
				end
			end
		end
	
	end
	
	self.styled = true
end

hooksecurefunc('ActionButton_Update', bBars_setStyle)

hooksecurefunc('PetActionBar_Update', function()
	if InCombatLockdown() then return end
	for i=1, NUM_PET_ACTION_SLOTS do
		local button  = _G['PetActionButton'..i]
		--button.styled = false
		bBars_setStyle(button)
	end  
end)


local bBars_UpdateAll = function()
	if InCombatLockdown() then return end
	for _, bar in pairs(bBars_frames) do
		local bar_name = bar:GetName()
		if _G[bar_name] then
		
			local opt = bBarsDB[pname][bar_name]
			
			if bar.GetName and bar_name ~= 'MultiCastActionBarFrame' then
				bar:SetSize(opt.size, opt.size)
				
				local cols = min(bBarsDB[pname][bar_name].col, #bar.buttons)
				local rows = ceil(#bar.buttons / (cols == 0 and 1 or cols))
				local spacing, padding = opt.gap, opt.gap
				local size = opt.size + spacing
				
				for index, button in pairs(bar.buttons) do
					local button_name = button:GetName()
					local button_shine = _G[button_name..'Shine']
					local button_autocastable = _G[button_name..'AutoCastable']
					local nt = _G[button_name..'NormalTexture'] and _G[button_name..'NormalTexture'] or button:GetNormalTexture() and button:GetNormalTexture() or nil
	
					if button_shine then button_shine:SetSize(opt.size, opt.size) end
					if button_autocastable then button_autocastable:SetSize((66/36) * opt.size, (66/36) * opt.size) end
					if nt then nt:SetSize((66/36) * opt.size, (66/36) * opt.size) end
					
					button:SetSize(opt.size, opt.size)

					if LBF and LBFGroup and nt and not lower(button_name):find('micro') then
						LBFGroup:RemoveButton(button)
						LBFGroup:AddButton(button)
						button.styled = true
					end
					
					button:ClearAllPoints()
					button:SetPoint('TOPLEFT', size*((index-1) % cols) + padding/2, -(size*(ceil(index/cols) - 1) + padding/2))
				end
				bar:SetSize(max(size*cols - spacing + padding*2, 4)-spacing, max(size*ceil(#bar.buttons/(cols==0 and 1 or cols)) - spacing + padding, 4))
			end
			
			if bar.updatefunc then bar.updatefunc(bar) end
			
			if not bBars_config then
				if bar.statedriver then 
					RegisterStateDriver(bar, 'visibility', bBarsDB[pname][bar_name].hide and 'hide' or bar.statedriver) 
				else
					RegisterStateDriver(bar, 'visibility', bBarsDB[pname][bar_name].hide and 'hide' or 'show')
				end
			end

			local point, relativeTo, relativePoint, x, y = unpack(opt.setPoint)
			bar:ClearAllPoints()
			if _G[relativeTo] then
				bar:SetPoint(point, relativeTo, relativePoint, x, y)
			else
				bar:SetPoint(point, 'bBars', relativePoint, x, y)
			end
		end
	end
	
	UIPARENT_MANAGED_FRAME_POSITIONS['CONTAINER_OFFSET_X'] = {baseX = 0, rightLeft = 2*32+3, rightRight = 32+3, isVar = 'xAxis'}
	MultiActionBar_Update()
    UIParent_ManageFramePositions()
end

local bBarsOptionsFrame= CreateFrame('frame', nil, UIParent)
bBarsOptionsFrame:RegisterEvent('VARIABLES_LOADED')
local bBarsOptionsPanel = CreateFrame('frame', 'bUnitFramesOptions')
bBarsOptionsPanel.name = 'bBars'

local bBars_MakeDropDown = function(self)
	bBarsOptionsPanel.dropdown = bBarsOptionsPanel.dropdown or CreateFrame('frame', '$parentDropDown', bBarsOptionsPanel, 'UIDropDownMenuTemplate')
	bBarsOptionsPanel.dropdown:SetPoint('TOPLEFT', 20, -40)
	bBarsOptionsPanel.dropdown:SetID(1)
	UIDropDownMenu_Initialize(bBarsOptionsPanel.dropdown, function()
		for v in pairs(bBarsDB) do UIDropDownMenu_AddButton({ text = v, func = function(self) UIDropDownMenu_SetSelectedID(bBarsOptionsPanel.dropdown, self:GetID()) bBarsOptionsPanel.dropdown.selectedValue = v end, checked = nil}) end
	end)
	UIDropDownMenu_SetSelectedID(bBarsOptionsPanel.dropdown, 1)
	bBarsOptionsPanel.dropdown.selectedValue = pname
end

bBarsOptionsFrame:SetScript('OnEvent', function(self)
	bBars_MakeDropDown(self)
	local title = bBarsOptionsPanel:CreateFontString('$parentTitle', 'OVERLAY', 'GameFontHighlight')
	title:SetText('Profiles')
	title:SetPoint('TOPLEFT', 40, -27)
	local b = CreateFrame('button', '$parentButton', bBarsOptionsPanel, 'UIPanelButtonTemplate')
	b:SetSize(80, 22)
	b:SetPoint('TOPLEFT', 170, -43)
	b:SetText('Use profile')
	b:SetScript('OnClick', function() bBarsDB[pname] = bBarsDB[UIDropDownMenu_GetSelectedValue(bBarsOptionsPanel.dropdown)] bBars_UpdateAll() end)
	
	InterfaceOptions_AddCategory(bBarsOptionsPanel)
	self:UnregisterEvent('VARIABLES_LOADED')
end)

local bBars_AddConfigButtons = function(f)
	local fname = f:GetName()
	if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
	
	if bBars_default_positions[fname] then
		if bBarsDB[pname][fname].col == nil then bBarsDB[pname][fname].col = bBars_default_positions[fname].col ~= nil and bBars_default_positions[fname].col or 12 end
		if bBarsDB[pname][fname].size == nil then bBarsDB[pname][fname].size = bBars_default_positions[fname].size ~= nil and bBars_default_positions[fname].size or 30 end
		if bBarsDB[pname][fname].gap == nil then bBarsDB[pname][fname].gap = bBars_default_positions[fname].gap ~= nil and bBars_default_positions[fname].gap or 1 end
		if bBarsDB[pname][fname].hide == nil then bBarsDB[pname][fname].hide = bBars_default_positions[fname].hide ~= nil and bBars_default_positions[fname].hide == 1 or false end
	else
		if bBarsDB[pname][fname].col == nil then bBarsDB[pname][fname].col = 12 end
		if bBarsDB[pname][fname].size == nil then bBarsDB[pname][fname].size = 30 end
		if bBarsDB[pname][fname].gap == nil then bBarsDB[pname][fname].gap = 1 end
		if bBarsDB[pname][fname].hide == nil then bBarsDB[pname][fname].hide = false end
	end
	
	if bBarsDB[pname][fname].setPoint == nil then
		if fname and bBars_default_positions[fname] then
			local opt = bBars_default_positions[fname]
			f:SetPoint(opt.a, opt.p, opt.pa, opt.x, opt.y)
		else
			f:SetPoint('CENTER', 0, 0)
		end
		
		local point, relativeTo, relativePoint, x, y = f:GetPoint()
		bBarsDB[pname][fname].setPoint = { point, relativeTo:GetName(), relativePoint, x, y }
	end

	f:SetScript('OnMouseWheel', function(self, value)
		if not bBars_config then return end
		if InCombatLockdown() or UnitAffectingCombat('player') then return end
		if IsShiftKeyDown() then
			if value > 0 then
				bBarsDB[pname][fname].size = bBarsDB[pname][fname].size + 1
			else
				bBarsDB[pname][fname].size = bBarsDB[pname][fname].size - 1
			end
		elseif IsAltKeyDown() then
			if value > 0 then
				bBarsDB[pname][fname].gap = bBarsDB[pname][fname].gap + 1
			else
				bBarsDB[pname][fname].gap = bBarsDB[pname][fname].gap - 1
			end
		else
			if value > 0 then
				if bBarsDB[pname][fname].col > 1 then bBarsDB[pname][fname].col = bBarsDB[pname][fname].col - 1 end
			else
				if bBarsDB[pname][fname].col < 12 then bBarsDB[pname][fname].col = bBarsDB[pname][fname].col + 1 end
			end
		end
		
		bBars_UpdateAll()
	end)
	
	if f.buttons ~= nil then
		for i,v in pairs(f.buttons) do
			local nt = _G[v:GetName()..'NormalTexture'] and _G[v:GetName()..'NormalTexture'] or v:GetNormalTexture() and v:GetNormalTexture() or nil
			if(LBF and LBFGroup and nt and not lower(v:GetName()):find('micro')) then 
				LBFGroup:AddButton(v)
				v.styled = true
			end
		end
	end
end

local bBars_getBar = function()
	local spec, vehicle
	if IsPandaria then spec = GetSpecialization() end		
	if IsPandaria then vehicle = '[vehicleui] 12' else vehicle = '[bonusbar:5] 11' end		
	local Page = {
		['DRUID'] = '[bonusbar:1,nostealth] 7; [bonusbar:1,stealth] 8; [bonusbar:2] 8; [bonusbar:3] 9; [bonusbar:4] 10;',		
		['WARRIOR'] = '[bonusbar:1] 7; [bonusbar:2] 8; [bonusbar:3] 9;',
		['PRIEST'] = '[bonusbar:1] 7;',
		['ROGUE'] = '[form:1][form:2/3,stealth] 7; [form:2/3,nostealth] 2;',
		['WARLOCK'] = '[form:2] 10;',
		['MONK'] = '[form:1] '..(spec == 1 and 8 or spec == 2 and 9 or spec == 3 and 7 or 9)..'; [form:2] 7;',
		['DEFAULT'] = vehicle..'; [bar:2] 2; [bar:3] 3; [bar:4] 4; [bar:5] 5; [bar:6] 6;',
	}
	local condition = Page['DEFAULT']
	local page = Page[class]
	if page then condition = condition..' '..page end
	condition = condition..' 1'
	return condition
end

local bBars_MakeActionButtonsFrame = function(name, id)
	local buttonId = 1
	local f = _G[name] or CreateFrame('frame', name, bBars, 'SecureHandlerStateTemplate')
	
	if id > 2 then f:SetAttribute('actionpage', id) end
	
	bBars_createTitle(f)
	
	f.buttons = {}
	
	for i=id*12-11, id*12 do
		local bname = name..'Button'..i
		local b, bind
		local modact = 1+(i-1)%12
		if id == 1 then
			--b = _G[bname] or CreateFrame('CheckButton', bname, f, 'ActionBarButtonTemplate')
			b = _G['ActionButton' .. i]
			if b then 
				f:SetFrameRef(b:GetName(), b)
				b:SetAttribute('binding', 'ACTIONBUTTON'..modact)
				b:SetAttribute("useparent-actionpage", true)
			end
		else
			if i <= 36 then
				b = _G[bname] or CreateFrame('CheckButton', bname, f, 'ActionBarButtonTemplate') 
				if b then b:SetAttribute('binding', 'MULTIACTIONBAR3BUTTON'..modact)  end
			elseif i <= 48 then
				b = _G[bname] or CreateFrame('CheckButton', bname, f, 'ActionBarButtonTemplate') 
				if b then b:SetAttribute('binding', 'MULTIACTIONBAR4BUTTON'..modact) end
			elseif i <= 60 then
				b = _G[bname] or CreateFrame('CheckButton', bname, f, 'ActionBarButtonTemplate') 
				if b then b:SetAttribute('binding', 'MULTIACTIONBAR2BUTTON'..modact) end
			elseif i <= 72 then
				b = _G[bname] or CreateFrame('CheckButton', bname, f, 'ActionBarButtonTemplate') 
				if b then b:SetAttribute('binding', 'MULTIACTIONBAR1BUTTON'..modact) end
			end
			
			if not b then
				--return
				-- for testing only!
				b = _G[bname] or CreateFrame('CheckButton', bname, f, 'ActionBarButtonTemplate') 
				b:SetAttribute('binding', 'MULTIACTIONBAR11BUTTON'..modact)
			end
		end

		if b then
			ActionBarButtonEventsFrame_OnLoad(b)
			b:SetScript('OnEvent', ActionButton_OnEvent)
		
			--if GetBindingKey(b:GetAttribute('binding')) then
				--SetOverrideBinding(b, true, GetBindingKey(b:GetAttribute('binding')), 'CLICK '..bname..':LeftButton')
			--end
			ActionButton_OnLoad(b)
			b:RegisterForClicks("AnyDown")
			b.RegisterForClicks = function() end
			b.actionbutton = true
			bBars_setStyle(b)
			b:SetParent(f)
			b:SetAttribute('unit2', 'player')
			--b:SetAttribute('showgrid', '1')
			
			bBars_UpdateBind(b)
			
			b.icon = _G[b:GetName()..'Icon']
			b:SetID(buttonId)
			buttonId = buttonId + 1
			
			if IsPandaria then ActionButton_UpdateAction(b) end
			
			hooksecurefunc(_G[b:GetName()..'HotKey'], 'SetVertexColor', function(self, ...)
				if InCombatLockdown() then return end
				local isUsable, notEnoughMana = IsUsableAction(b.action)
				if (ActionHasRange(b.action) and IsActionInRange(b.action) == 0) then
					b.icon:SetVertexColor(1, 0, 0)
					return
				elseif (notEnoughMana) then
					b.icon:SetVertexColor(.2, .2, .2)
					return
				elseif (isUsable) then
					b.icon:SetVertexColor(1, 1, 1, 1)
					return
				else
					b.icon:SetVertexColor(0.4, 0.4, 0.4, 1)
					return
				end			
			end)
			
			b:Show()
			b.Hide = b.Show
			
			if id == 1 then
				b:SetAttribute('_childupdate-page', [[ self:SetAttribute('actionpage', message) ]])
			end			
			
			tput(f.buttons, b)
		end
	end
	
	if id == 1 then
		--f:Execute([[ buttons = table.new() for i = 1, 12 do table.insert(buttons, self:GetFrameRef('bMultiBar1Button'..i)) end ]])
		f:SetAttribute('_onstate-page', [[ control:ChildUpdate('page', newstate) ]])
		RegisterStateDriver(f, 'page', bBars_getBar())
		if petbattle ~= '' then
			f.statedriver = petbattle..' hide; show'
		end
	else
		f.statedriver = '[vehicleui]'..petbattle..' hide; show'
	end
	
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local bBars_MakeTotemFrame = function()
	if (class == 'SHAMAN') then
		local f = MultiCastActionBarFrame
		local fname = f:GetName()
		
		bBars_createTitle(f)
		
		if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
		if bBarsDB[pname][fname].col == nil then bBarsDB[pname][fname].col = 6 end
		if bBarsDB[pname][fname].size == nil then bBarsDB[pname][fname].size = 30 end
		
		UIPARENT_MANAGED_FRAME_POSITIONS['f'] = nil
		f:SetScript('OnUpdate', nil)
		f_OnUpdate = nullfunc
		f:SetScript('OnShow', nil)
		f:SetScript('OnHide', nil)
		f:SetParent(bBars)
		f:ClearAllPoints()
		f:SetPoint('CENTER')
		MultiCastFlyoutFrame:SetFrameStrata('HIGH')
		f.SetParent = nullfunc
		MultiCastRecallSpellButton.SetPoint = nullfunc
		if not f.buttons then
			f.buttons = {}
			for i = 1, NUM_MULTI_CAST_BUTTONS_PER_PAGE do
				tput(f.buttons, _G['MultiCastSlotButton'..i])
			end
			tput(f.buttons, _G['MultiCastSummonSpellButton'])
			tput(f.buttons, _G['MultiCastRecallSpellButton'])
		end
		
		f.statedriver = '[vehicleui]'..petbattle..' hide; show'
		bBars_AddConfigButtons(f)
		tput(bBars_frames, f)
	end
end

local bBars_WatchFrame = function()
	local fname = 'bWatchFrame'
	local f = _G[fname] or CreateFrame('frame', fname, bBars)
	f:SetSize(270, 250)
	f.SetSize = function() end
	f:SetClampedToScreen(true)
	
	bBars_createTitle(f)
	
	if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
	if bBarsDB[pname][fname].col == nil then bBarsDB[pname][fname].col = 6 end
	if bBarsDB[pname][fname].size == nil then bBarsDB[pname][fname].size = 30 end
	if not f.buttons then f.buttons = {} end

	WatchFrame:SetParent(f)
	WatchFrame:SetFrameStrata("LOW")
	WatchFrame:SetFrameLevel(3)
	WatchFrame:SetClampedToScreen(false)
	WatchFrame:ClearAllPoints()
	WatchFrame.ClearAllPoints = function() end
	WatchFrame:SetAllPoints()
	WatchFrame.GetTop = function() return 250 end
	WatchFrame.GetBottom = function() return 1 end
	WatchFrame.SetPoint = function() end

	WatchFrameTitle:SetParent(f)
	WatchFrameCollapseExpandButton:SetParent(f)
	WatchFrameCollapseExpandButton:SetSize(16, 16)
	WatchFrameCollapseExpandButton:SetFrameStrata(WatchFrameHeader:GetFrameStrata())
	WatchFrameCollapseExpandButton:SetFrameLevel(WatchFrameHeader:GetFrameLevel() + 1)
	
	bBars_AddConfigButtons(f)
	WatchFrame.statedriver = '[vehicleui]'..petbattle..' hide; show'
	tput(bBars_frames, f)
end

local bBars_updaterepbar = function(self)
	self:SetSize(bBarsDB[pname][self:GetName()].size*2, 12)
	local name, reaction, minv, maxv, value = GetWatchedFactionInfo()
	if name then
		self:SetMinMaxValues(minv, maxv)
		self:SetValue(value)
		maxv = maxv - minv
		value = value - minv
		local color = FACTION_BAR_COLORS[reaction]
		self:SetStatusBarColor(color.r, color.g, color.b)
		self.txt:SetText(name..' '..value..'/'..maxv)
	else
		self:SetMinMaxValues(0,0)
		self:SetValue(0)
		self.txt:SetText()
	end
	
	local divWidth = self:GetWidth()/20
	local xpos = divWidth
	for i=1, 19 do
		local texture = _G['repbardiv'..i] or self:CreateTexture('repbardiv'..i, 'OVERLAY') texture:SetTexture(1, .8, .1) texture:SetSize(1, self:GetHeight()) texture:SetVertexColor(0,0,0)
		local xalign = floor(xpos)
		texture:SetPoint('LEFT', xalign, 0)
		xpos = xpos + divWidth
	end
end

local bBars_MakeRepBarFrame = function()
	local fname = 'bRepBarFrame'
	if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
	if bBarsDB[pname][fname].col == nil then bBarsDB[pname][fname].col = 6 end
	if bBarsDB[pname][fname].size == nil then bBarsDB[pname][fname].size = 200 end
	if bBarsDB[pname][fname].gap == nil then bBarsDB[pname][fname].gap = 10 end
	if bBarsDB[pname][fname].hide == nil then bBarsDB[pname][fname].hide = 0 end
	
    local f = _G[fname] or CreateFrame('StatusBar', fname, bBars)
	
	bBars_createTitle(f)
	
	f.bg = f.bg or f:CreateTexture(nil, 'BACKGROUND')
	f.bg:SetAllPoints()
	f.bg:SetTexture(backgroundtex)
	f.bg:SetVertexColor(0,0,0,.7)
	
    f:SetStatusBarTexture(statusbartex)
    f:SetStatusBarColor(0, .8, 0)
    f:RegisterEvent('CVAR_UPDATE')
    f:RegisterEvent('UPDATE_FACTION')
    f:RegisterEvent('PLAYER_ENTERING_WORLD')

	f.txt = f.txt or f:CreateFontString(nil, 'OVERLAY', 'GameFontHighlightSmall')
	f.txt:SetPoint('CENTER', 0, -2)
	
	f.updatefunc = f.updatefunc or bBars_updaterepbar
	f:SetScript('OnEvent', f.updatefunc)
	bBars_updaterepbar(f)
	
	f.buttons = {}
	
	f.statedriver = '[vehicleui]'..petbattle..' hide; show'
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

SECOND_NUMBER_CAP = SECOND_NUMBER_CAP:lower()
FIRST_NUMBER_CAP = FIRST_NUMBER_CAP:lower()
SetCVar("breakUpLargeNumbers", 1)
local fixvalue = function(val)
   if(val >= 1e6) then
	  return ('%.2f'..SECOND_NUMBER_CAP):format(val / 1e6):gsub('%.?0+(['..FIRST_NUMBER_CAP..SECOND_NUMBER_CAP..'])$', '%1')
   elseif(val >= 1e4) then
	  return ('%.1f'..FIRST_NUMBER_CAP):format(val / 1e3):gsub('%.?0+(['..FIRST_NUMBER_CAP..SECOND_NUMBER_CAP..'])$', '%1')
   else
	  return val
   end
end

local bBars_updateexpbar = function(self)
	self:SetSize(bBarsDB[pname][self:GetName()].size*2, 12)
	
	self.bg:ClearAllPoints()
	self.bg:SetPoint('TOPLEFT', -1, 1)
	self.bg:SetPoint('BOTTOMRIGHT', 1, -1)		
	
	local currXP, nextXP = UnitXP('player'), UnitXPMax('player')
	self:SetMinMaxValues(math.min(0, currXP), nextXP)
	self:SetValue(currXP)
	local treshold = GetXPExhaustion()
	local eID, eName, eMult = GetRestState()
	if (eID and eID >= 3) then self.arrow:SetPoint('CENTER', self, 'RIGHT', 0, 0) end		
	
	self.txt:SetText(format('%s/%s (%s)|r', fixvalue(UnitXP('player') + 1), fixvalue(UnitXPMax('player')-UnitXP('player') + 1), fixvalue(GetXPExhaustion() or 0)))
	
	if not treshold then
		self.arrow:Hide() self.more:Hide()
	else
		local tick = math.max(((currXP + treshold) / nextXP) * self:GetWidth(), 0)
		self.arrow:ClearAllPoints()
		
		if (tick > self:GetWidth()) then
			self.arrow:SetPoint('CENTER', self, 'RIGHT', 0, 0)
			self.more:SetPoint('TOPRIGHT', self, 'TOPRIGHT', 0, 0)
		else
			self.arrow:Show()
			self.arrow:SetPoint('CENTER', self, 'LEFT', Fix(tick), 0)
			self.more:Show()
			self.more:SetPoint('TOPRIGHT', self, 'TOPLEFT', Fix(tick), 0)
		end
	end		

	local RestState = GetRestState()
	if RestState == 1 then
		self:SetStatusBarColor(0, .4, .9)
		self.more:SetVertexColor(0, .4, .9, .3)
		self.arrow:GetHighlightTexture():SetVertexColor(0, .4, .9)
	elseif RestState == 2 then
		self:SetStatusBarColor(.6, 0, .5)
		self.more:SetVertexColor(.6, 0, .5, .3)
		self.arrow:GetHighlightTexture():SetVertexColor(.6, 0, .5)
	end
	
	local divWidth = self:GetWidth()/20
	local xpos = divWidth
	for i=1, 19 do
		local texture = _G['expbardiv'..i] or self:CreateTexture('expbardiv'..i, 'OVERLAY') texture:SetTexture(1, .8, .1) texture:SetSize(1, self:GetHeight()) texture:SetVertexColor(0,0,0)
		local xalign = floor(xpos)
		texture:SetPoint('LEFT', xalign, 0)
		xpos = xpos + divWidth
	end
end

local bBars_MakeExpBarFrame = function()	
	if (not IsPandaria and UnitLevel('player') == 85) or (IsPandaria and UnitLevel('player') == 90) then return end
	
	local fname = 'bExpBarFrame'
	if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
	if bBarsDB[pname][fname].col == nil then bBarsDB[pname][fname].col = 6 end
	if bBarsDB[pname][fname].size == nil then bBarsDB[pname][fname].size = 200 end
	if bBarsDB[pname][fname].gap == nil then bBarsDB[pname][fname].gap = 10 end
	if bBarsDB[pname][fname].hide == nil then bBarsDB[pname][fname].hide = 0 end
	
	local f = _G[fname] or CreateFrame('StatusBar', fname, bBars)
	
	f.bg = f.bg or f:CreateTexture(nil, 'BORDER')
	f.bg:SetPoint('TOPLEFT', -1, 1)
	f.bg:SetPoint('BOTTOMRIGHT', 1, -1)
	f.bg:SetTexture(backgroundtex)
	f.bg:SetVertexColor(0,0,0,.7)
	
	f.more = f.more or f:CreateTexture(nil, 'OVERLAY')
	f.more:SetPoint('TOPLEFT', 0, 0)
	f.more:SetSize(0, 18)
	f.more:SetTexture(1,1,1,1)
	f.more:SetBlendMode('ADD')
	
	f.arrow = f.arrow or CreateFrame('Button', nil, f)
	f.arrow:SetSize(32,32)
	f.arrow:SetPoint('CENTER')
	f.arrow:SetNormalTexture('Interface\\MainMenuBar\\UI-ExhaustionTickNormal')
	f.arrow:SetHighlightTexture('Interface\\MainMenuBar\\UI-ExhaustionTickHighlight')
	f.arrow:GetHighlightTexture():SetBlendMode('ADD')
	f.arrow:SetScript('OnLeave', GameTooltip_Hide)
	f.arrow:SetScript('OnEnter', ExhaustionToolTipText)
	
	f.txt = f.txt or f:CreateFontString(nil, 'OVERLAY')
	f.txt:SetFont(bBars_font, fontSize, fontStyle)
	f.txt:SetPoint('CENTER', 0, 0)

	f:SetStatusBarTexture(statusbartex)
	f:SetStatusBarColor(.1, .2, .8)
	f:RegisterEvent('PLAYER_ENTERING_WORLD')
	f:RegisterEvent('PLAYER_XP_UPDATE')
	f:RegisterEvent('UPDATE_EXHAUSTION')
	f:RegisterEvent('PLAYER_LEVEL_UP')
	f:RegisterEvent('PLAYER_UPDATE_RESTING')
	f.updatefunc = f.updatefunc or bBars_updateexpbar
	f:SetScript('OnEvent', f.updatefunc)
	f.updatefunc(f)
	
	f.buttons = {}
	
	f.statedriver = '[vehicleui]'..petbattle..' hide; show'
	
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local bBars_MakePetFrame = function()
	local fname = 'bPetFrame'
	local f = _G[fname] or CreateFrame('frame', fname, bBars, 'SecureHandlerStateTemplate')
	
	bBars_createTitle(f)
	
	f.buttons = {}
	
	for i=1, 12 do
		local b = _G['PetActionButton'..i]
		
		if b then
			b:SetAttribute('binding', 'BONUSACTIONBUTTON'..i)
			b:SetParent(f)
			b:SetID(i)
			bBars_setStyle(b)
			bBars_UpdateBind(b)
			tput(f.buttons, b)
		end
	end
	
	f.statedriver = '[vehicleui]'..petbattle..' hide; show'
	
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local bBars_shapeshiftUpdate = function(self)
	local texture, name, isActive, isCastable = GetShapeshiftFormInfo(self:GetID())
	self:SetChecked(isActive)
	self.icon:SetTexture(texture)
	if texture then
		local start, duration, enable = GetShapeshiftFormCooldown(self:GetID())
		CooldownFrame_SetTimer(self.cooldown, start, duration, enable)
	else
		self.cooldown:Hide()
	end
end

local bBars_MakeShapeShiftFrame = function()
	local fname = 'bShapeshiftFrame'
	
	local f = _G[fname] or CreateFrame('frame', fname, bBars, 'SecureHandlerStateTemplate')
	
	bBars_createTitle(f)
	
	f.buttons = {}
	
	for i = 1, 10 do
		local bname = 'bShapeshiftButton'..i
		local b = _G[bname] or CreateFrame('CheckButton', bname, f, 'StanceButtonTemplate')
		b:SetID(i)
		b:RegisterForClicks('AnyUp')
		
		b.icon = b.icon or _G[bname..'Icon'] or b:CreateTexture(bname..'Icon', 'BACKGROUND')
		b.icon:SetPoint('TOPLEFT', 1, -1)
		b.icon:SetPoint('BOTTOMRIGHT', -1, 1)
		b.icon:SetDrawLayer('BACKGROUND', 1)
		
		b:SetScript('OnEnter', function(self) GameTooltip:SetOwner(self, 'ANCHOR_RIGHT') GameTooltip:SetShapeshift(self:GetID()) end)
		b:SetScript('OnLeave', function(self) GameTooltip:Hide() end)
		
		b.hotkey = b.hotkey or _G[bname..'HotKey'] or b:CreateFontString(bname..'HotKey', 'ARTWORK')
		b.hotkey:SetFont(bBars_font, fontSize, fontStyle)
		b.hotkey:SetPoint('TOPRIGHT', 2, -2)
		b.hotkey:SetJustifyH('RIGHT')
		b.hotkey:SetHeight(10)

		b.cooldown = b.cooldown or _G[bname..'Cooldown'] or CreateFrame('Cooldown', bname..'Cooldown', b, 'CooldownFrameTemplate')
		b.cooldown:SetAllPoints(b)
		
		b:SetAttribute('type', 'spell')
		local texture, name = GetShapeshiftFormInfo(b:GetID())
		b:SetAttribute('spell', name)
		b.icon:SetTexture(texture)
		
		b:RegisterEvent('UPDATE_BINDINGS')
		b:RegisterEvent('UPDATE_SHAPESHIFT_FORMS')
		b:RegisterEvent('UPDATE_SHAPESHIFT_COOLDOWN')
		b:RegisterEvent('PLAYER_ENTERING_WORLD')
		b:RegisterEvent('SPELL_UPDATE_COOLDOWN')
		b:RegisterEvent('SPELL_UPDATE_USABLE')
		b:RegisterEvent('UNIT_AURA')
		b:RegisterEvent('PLAYER_TALENT_UPDATE')
		
		b:SetScript('OnEvent', function(self, event, arg1)
			if not InCombatLockdown() then
				if self:GetID() > GetNumShapeshiftForms() then self:Hide() else self:Show() end
			end
			
			if event == 'UPDATE_BINDINGS' then
				bBars_UpdateBind(self)
			elseif event == 'UNIT_AURA' then
				if arg1 == 'player' then
					bBars_shapeshiftUpdate(self)
				end
			elseif event == 'UPDATE_SHAPESHIFT_COOLDOWN' then
				bBars_shapeshiftUpdate(self)
			elseif event == 'PLAYER_TALENT_UPDATE' then
				if not InCombatLockdown() then
					local texture, name = GetShapeshiftFormInfo(self:GetID())
					self:SetAttribute('type', 'spell')
					self:SetAttribute('spell', name)
					self.icon:SetTexture(texture)
				end
			else
				bBars_shapeshiftUpdate(self)
			end
		end)
		
		b:SetNormalTexture([[Interface\Buttons\UI-Quickslot2]])
		b:SetPushedTexture([[Interface\Buttons\UI-Quickslot-Depress]])
		b:SetHighlightTexture([[Interface\Buttons\ButtonHilight-Square]])
		b:SetCheckedTexture([[Interface\Buttons\CheckButtonHilight]])

		if i > GetNumShapeshiftForms() then b:Hide() end
		bBars_setStyle(b)
		
		tput(f.buttons, b)
	end
	
	f.statedriver = (class == 'MAGE' or class == 'SHAMAN') and 'hide' or '[vehicleui]'..petbattle..' hide; show'
	
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local bBars_updatepetexpbar = function(self, ...)
	self:SetSize(bBarsDB[pname][self:GetName()].size*2, 10)
	local xp, maxXp = C_PetBattles.GetXP(LE_BATTLE_PET_ALLY, C_PetBattles.GetActivePet(LE_BATTLE_PET_ALLY))
	self:SetMinMaxValues(0, maxXp)
	self:SetValue(xp)
	self.txt:SetFormattedText(PET_BATTLE_CURRENT_XP_FORMAT, xp, maxXp)
	
	local divWidth = self:GetWidth()/20
	local xpos = divWidth
	for i=1, 19 do
		local texture = _G['bpetexpbardiv'..i] or self:CreateTexture('bpetexpbardiv'..i, 'OVERLAY') texture:SetTexture(1, .8, .1) texture:SetSize(1, self:GetHeight()) texture:SetVertexColor(0,0,0)
		local xalign = floor(xpos)
		texture:SetPoint('LEFT', xalign, 0)
		xpos = xpos + divWidth
	end
end

local bBars_MakeBattlePetExpBarFrame = function()
	if not IsPandaria then return end
	
	local fname = 'bBattlePetExpBarFrame'
	if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
	if bBarsDB[pname][fname].col == nil then bBarsDB[pname][fname].col = 6 end
	if bBarsDB[pname][fname].size == nil then bBarsDB[pname][fname].size = 117 end
	if bBarsDB[pname][fname].gap == nil then bBarsDB[pname][fname].gap = 10 end
	
	local f = _G[fname] or CreateFrame('StatusBar', fname, bBars)
	
	bBars_createTitle(f)
	
	f.bg = f.bg or f:CreateTexture(nil, 'BORDER')
	f.bg:SetPoint('TOPLEFT', -1, 1)
	f.bg:SetPoint('BOTTOMRIGHT', 1, -1)
	f.bg:SetTexture(backgroundtex)
	f.bg:SetVertexColor(1,1,1,.7)

	f.txt = f.txt or f:CreateFontString(nil, 'OVERLAY', 'GameFontHighlightSmall')
	f.txt:SetPoint('CENTER')
	
	f:SetStatusBarTexture(statusbartex)
	f:SetStatusBarColor(.1, .2, .8)
	f:RegisterEvent('PET_BATTLE_HEALTH_CHANGED')
	f:RegisterEvent('PET_BATTLE_HEALTH_CHANGED')
	f:RegisterEvent('PET_BATTLE_PET_CHANGED')
	f:RegisterEvent('PET_BATTLE_CLOSE')
	f.updatefunc = f.updatefunc or bBars_updatepetexpbar
	f:SetScript('OnEvent', f.updatefunc)
	
	f.buttons = {}
	bBars_AddConfigButtons(f)
	f.statedriver = petbattle..' show; hide'
	tput(bBars_frames, f)
end

local bBars_MakePetBattleFrame = function()
	if not IsPandaria then return end
	local fname = 'bPetBattleFrame'
	local f = _G[fname] or CreateFrame('frame', fname, bBars, 'SecureHandlerStateTemplate')
	
	bBars_createTitle(f)
	
	f.buttons = {}
	
	f:RegisterEvent('PET_BATTLE_PET_CHANGED')
	f:SetScript('OnEvent', function(self, event)
		for i=1, NUM_BATTLE_PET_ABILITIES do
			PetBattleAbilityButton_UpdateIcons(self.buttons[i])
			PetBattleActionButton_UpdateState(self.buttons[i])
		end
		PetBattleActionButton_UpdateState(f.SwitchPetButton)
		PetBattleActionButton_UpdateState(f.CatchButton)
		PetBattleActionButton_UpdateState(f.SkipButton)
	end)
	
	for i=1, NUM_BATTLE_PET_ABILITIES do
		local bname = 'PetBattleButton'..i
		local b = _G[bname] or CreateFrame('CheckButton', bname, f, 'PetBattleAbilityButtonTemplate', i)
		b:SetID(i)
		bBars_setStyle(b)
		tput(f.buttons, b)
	end
	
	f.SkipButton = f.SkipButton or CreateFrame('CheckButton', 'bSkipButton', f, 'PetBattleActionButtonTemplate')
	PetBattleActionButton_Initialize(f.SkipButton, LE_BATTLE_PET_ACTION_SKIP)
	f.SkipButton.Icon:SetTexture("Interface\\Icons\\misc_arrowright")
	f.SkipButton:SetScript('OnClick', function() C_PetBattles.SkipTurn() end)
	f.SkipButton:SetScript('OnEnter', function(self) GameTooltip:SetOwner(self, 'ANCHOR_RIGHT') GameTooltip:AddLine(PET_BATTLE_PASS) GameTooltip:Show() end)
	f.SkipButton:SetScript('OnLeave', function(self) GameTooltip:Hide() end)
	
	bBars_setStyle(f.SkipButton)
	tput(f.buttons, f.SkipButton)
	
	hooksecurefunc('PetBattleFrame_Remove', function(self)
		PetBattleFrame.BottomFrame.PetSelectionFrame:Hide()
	end)
	
	f.SwitchPetButton = f.SwitchPetButton or CreateFrame('CheckButton', 'bSwitchPetButton', f, 'PetBattleActionButtonTemplate')
	f.SwitchPetButton.AdditionalIcon = f.SwitchPetButton.AdditionalIcon or f.SwitchPetButton:CreateTexture('AdditionalIcon', 'OVERLAY', 'BattleBar-SwapPetIcon')
	f.SwitchPetButton.AdditionalIcon:SetAllPoints()
	PetBattleActionButton_Initialize(f.SwitchPetButton, LE_BATTLE_PET_ACTION_SWITCH_PET)
	f.SwitchPetButton.Icon:SetTexture("Interface\\Icons\\INV_BOX_BIRDCAGE_01")
	f.SwitchPetButton:SetScript('OnClick', function(self)
		local battleState, selectedActionType = C_PetBattles.GetBattleState(), C_PetBattles.GetSelectedAction()
		local mustSwap = (( not selectedActionType) or (selectedActionType == BATTLE_PET_ACTION_NONE)) and (battleState == LE_PET_BATTLE_STATE_WAITING_PRE_BATTLE) or (battleState == LE_PET_BATTLE_STATE_WAITING_FOR_FRONT_PETS)
		if PetBattleFrame.BottomFrame.PetSelectionFrame:IsShown() and not mustSwap then
			PetBattlePetSelectionFrame_Hide(PetBattleFrame.BottomFrame.PetSelectionFrame)
		else
			PetBattlePetSelectionFrame_Show(PetBattleFrame.BottomFrame.PetSelectionFrame)
		end
	end)
	f.SwitchPetButton:SetScript('OnEnter', function(self) GameTooltip:SetOwner(self, 'ANCHOR_RIGHT') GameTooltip:AddLine(PET_BATTLE_SELECT_A_PET) GameTooltip:Show() end)
	f.SwitchPetButton:SetScript('OnLeave', function(self) GameTooltip:Hide() end)
	bBars_setStyle(f.SwitchPetButton)
	tput(f.buttons, f.SwitchPetButton)
	
	f.CatchButton = f.CatchButton or CreateFrame('CheckButton', 'bCatchButton', f, 'PetBattleActionButtonTemplate')
	PetBattleActionButton_Initialize(f.CatchButton, LE_BATTLE_PET_ACTION_TRAP)
	f.CatchButton.Icon:SetTexture("Interface\\Icons\\ABILITY_HUNTER_TRAPLAUNCHER")
	f.CatchButton:SetScript('OnClick', function() C_PetBattles.UseTrap() end)
	f.CatchButton:SetScript('OnEnter', function(self) GameTooltip:SetOwner(self, 'ANCHOR_RIGHT') GameTooltip:AddLine('Hook') GameTooltip:Show() end)
	f.CatchButton:SetScript('OnLeave', function(self) GameTooltip:Hide() end)
	bBars_setStyle(f.CatchButton)
	tput(f.buttons, f.CatchButton)
		
	f.ForfeitButton = f.ForfeitButton or CreateFrame('CheckButton', 'bForfeitButton', f, 'PetBattleActionButtonTemplate')
	f.ForfeitButton.Icon:SetTexture("Interface\\Icons\\ACHIEVEMENT_BG_GETXFLAGS_NO_DIE")
	f.ForfeitButton:SetScript('OnClick', function() C_PetBattles.ForfeitGame() end)
	f.ForfeitButton:SetScript('OnEnter', function(self) GameTooltip:SetOwner(self, 'ANCHOR_RIGHT') GameTooltip:AddLine(COMPLETE_QUEST) GameTooltip:Show() end)
	f.ForfeitButton:SetScript('OnLeave', function(self) GameTooltip:Hide() end)
	bBars_setStyle(f.ForfeitButton)
	tput(f.buttons, f.ForfeitButton)
	
	UpdateMicroButtonsParent = nullfunc
	MoveMicroButtons = nullfunc

	f.statedriver = petbattle..' show; hide'
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local bBars_MakeBagsFrame = function()
	local fname = 'bBagsFrame'
	
	local f = _G[fname] or CreateFrame('frame', fname, bBars)
	
	bBars_createTitle(f)
	
	f.buttons = {_G['CharacterBag3Slot'], _G['CharacterBag2Slot'], _G['CharacterBag1Slot'], _G['CharacterBag0Slot'], _G['MainMenuBarBackpackButton']}
	for i,v in pairs(f.buttons) do
		v:SetParent(f)
	end
	
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local getButtons = function(self)
	self.buttons = {}
	for i, v in pairs(MICRO_BUTTONS) do
		local b = _G[v]
		b:SetHitRectInsets(0, 0, 0, 0)
		local name = b:GetName()
		if _G[name..'Texture'] then
			_G[name..'Texture']:SetTexCoord(0, .65, 0, .65)
			_G[name..'Texture']:ClearAllPoints()
			_G[name..'Texture']:SetAllPoints()
		end
		if _G[name..'Tabard'] then
			_G[name..'Tabard']:ClearAllPoints()
			_G[name..'Tabard']:SetAllPoints()
			
			if _G[name..'TabardBackground'] then
				_G[name..'TabardBackground']:SetTexCoord(.1, .9, .45, .9)
				_G[name..'TabardBackground']:ClearAllPoints()
				_G[name..'TabardBackground']:SetAllPoints(v)
			end
			
			if _G[name..'TabardEmblem'] then
				_G[name..'TabardEmblem']:ClearAllPoints()
				_G[name..'TabardEmblem']:SetPoint('TOPLEFT', v, 4, -4)
				_G[name..'TabardEmblem']:SetPoint('BOTTOMRIGHT', v, -4, 4)
			end
		end
		
		if _G['MainMenuBarPerformanceBar'] then
			_G['MainMenuBarPerformanceBar']:SetTexture(0,0,0)
			_G['MainMenuBarPerformanceBar']:ClearAllPoints()
			_G['MainMenuBarPerformanceBar']:SetAllPoints()
			_G['MainMenuBarPerformanceBar']:SetAlpha(.1)
		end
		b:SetParent(self)
		b.SetParent = nullfunc
		bBars_setStyle(b)
		tput(self.buttons, b)
	end
	
	do
		MicroButtonPortrait:ClearAllPoints()
		MicroButtonPortrait:SetPoint('TOPLEFT', CharacterMicroButton, 'TOPLEFT', 2, -2)
		MicroButtonPortrait:SetPoint('BOTTOMRIGHT', CharacterMicroButton, 'BOTTOMRIGHT', -2, 2)
	end

	return self.buttons
end

local bBars_MakeMicroMenuFrame = function()
	local fname = 'bMicroMenuFrame'
	local f = _G[fname] or CreateFrame('frame', fname, bBars)
	
	bBars_createTitle(f)
	
	f.buttons = f.buttons or getButtons(f)
	
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local bBars_MakeVehicleFrame = function()
	local fname = 'bVehicleFrame'
	local f = _G[fname] or CreateFrame('frame', fname, bBars, 'SecureHandlerStateTemplate')
	
	bBars_createTitle(f)
	
	f.buttons = {}

	local b = _G['bVehicleLeaveButton'] or CreateFrame('Button', 'bVehicleLeaveButton', f)
	b:SetNormalTexture('Interface\\Vehicles\\UI-Vehicles-Button-Exit-Up')
	b:SetPushedTexture('Interface\\Vehicles\\UI-Vehicles-Button-Exit-Down')
	b:SetHighlightTexture('Interface\\Vehicles\\UI-Vehicles-Button-Exit-Down')
	
	b:RegisterEvent('UNIT_ENTERED_VEHICLE')
	b:RegisterEvent('UNIT_EXITED_VEHICLE')
	b:SetSize(32, 32)
	
	b:SetScript('OnClick', function(self) VehicleExit() end)
	b:SetScript('OnEvent', function(self) if CanExitVehicle() then self:Show() else self:Hide() end end)
	b:Hide()
	
	f.statedriver = '[vehicleui]show; '..petbattle..'[] hide'
	
	bBars_AddConfigButtons(f)
	tput(f.buttons, b)
	tput(bBars_frames, f)
end

local bBars_MakeExtraBarFrame = function()
	local fname = 'bExtraBarFrame'
	local f = _G[fname] or CreateFrame('frame', fname, bBars, 'SecureHandlerShowHideTemplate')
	
	bBars_createTitle(f)
	
	f.buttons = {}

	local i = 1
	while _G['ExtraActionButton'..i] ~= nil do
		local b = _G['ExtraActionButton'..i]
		b.style:SetTexture(nil)
		b:SetParent(f)
		
		tput(f.buttons, b)
		i = i + 1
	end
	
	f.statedriver = '[extrabar]show;'..petbattle..'[] hide'
	
	bBars_AddConfigButtons(f)
	tput(bBars_frames, f)
end

local bBars_ConfigBars = function(isConfig)
	if InCombatLockdown() or UnitAffectingCombat('player') then return end
	if isConfig == true then
		bBars_warning:Show()
		for i, frm in pairs(bBars_frames) do
			UnregisterStateDriver(frm, 'visibility')
			
			frm:Show()
			
			frm:EnableMouseWheel(1)
			frm:SetBackdrop({ bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background', })
			frm:SetBackdropColor(1, 1, 1)
			if bBarsDB[pname][frm:GetName()].hide then frm:SetBackdropColor(1, 0, 0, .3) else frm:SetBackdropColor(1, 1, 1, 1) end
			if frm.title then frm.title:Show() end
			
			for j,v in pairs(frm.buttons) do
				v:SetFrameLevel(frm:GetFrameLevel()-1)
				v:EnableMouse(false)
			end
			frm:EnableMouse(true)
			frm:SetScript('OnMouseDown', function(self, buttonkey) 
				if buttonkey == 'LeftButton' and not IsModifierKeyDown() then
					self:StartMoving()
				end
			end)
			frm:SetMovable(true)
			frm:SetResizable(true)
			frm:SetClampedToScreen(true)
			
			frm:SetScript('OnMouseUp', function(self, buttonkey)
				local fname = frm:GetName()
				if buttonkey == 'LeftButton' and not IsModifierKeyDown() then
					self:StopMovingOrSizing()
					for i,v in pairs(bBars_frames) do
						if v ~= self then
							if bBarsDB[pname].stick == 1 then
								bBars_Stick(self, v, 16)
							end
						end
					end
					
					local point, relativeTo, relativePoint, x, y = self:GetPoint()
					if relativeTo then
						bBarsDB[pname][fname].setPoint = { point, relativeTo:GetName(), relativePoint, x, y }
					else
						bBarsDB[pname][fname].setPoint = { point, 'bBars', relativePoint, x, y }
					end
				else
					if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
					if bBarsDB[pname][fname].hide then
						bBarsDB[pname][fname].hide = false
						self:SetBackdropColor(1, 1, 1, 1)
					else
						bBarsDB[pname][fname].hide = true
						self:SetBackdropColor(1, 0, 0, .3)
					end
				end
			end)
		end
	else
		bBars_warning:Hide()
		for i,frm in pairs(bBars_frames) do
			frm:SetBackdrop(nil)
			frm:EnableMouseWheel(0)
			
			for j,v in pairs(frm.buttons) do
				v:SetFrameLevel(frm:GetFrameLevel()+1)
				v:EnableMouse(true)
			end
			
			if frm.title then frm.title:Hide() end
			
			frm:SetMovable(false)
			frm:SetResizable(false)
			frm:EnableMouse(false)
			frm:SetScript('OnMouseDown', nil)
			frm:SetScript('OnMouseUp', nil)
			local fname = frm:GetName()
			if bBarsDB[pname][fname] == nil then bBarsDB[pname][fname] = {} end
			if bBarsDB[pname][fname].hide then frm:Hide() else frm:Show() end
			if frm.statedriver then
				RegisterStateDriver(frm, 'visibility', bBarsDB[pname][fname].hide and 'hide' or frm.statedriver)
			else
				RegisterStateDriver(frm, 'visibility', bBarsDB[pname][fname].hide and 'hide' or 'show')
			end
		end
	end
end

-- minimap button

local function MoveMinimapButton()
	local xpos, ypos
	local minimapShape = GetMinimapShape and GetMinimapShape() or 'ROUND'
	if minimapShape == 'SQUARE' then
		xpos = 110 * cos(bBarsDB[pname].minimapPos)
		ypos = 110 * sin(bBarsDB[pname].minimapPos)
		xpos = math.max(-82, math.min(xpos, 84))
		ypos = math.max(-86, math.min(ypos, 82))
	else
		xpos = 80 * cos(bBarsDB[pname].minimapPos)
		ypos = 80 * sin(bBarsDB[pname].minimapPos)
	end
	_G['bBarsMinimapButton']:SetPoint('TOPLEFT', Minimap, 'TOPLEFT', 54-xpos, ypos-54)
end

local function DragMinimapButton()
	local xpos, ypos = GetCursorPosition()
	local xmin, ymin = Minimap:GetLeft() or 400, Minimap:GetBottom() or 400
	local scale = Minimap:GetEffectiveScale()
	xpos = xmin-xpos/scale+70
	ypos = ypos/scale-ymin-70
	bBarsDB[pname].minimapPos = math.deg(math.atan2(ypos, xpos))
	MoveMinimapButton()
end

local bBars_CreateMinimapButton = function()
	if _G['bBarsMinimapButton'] then return end

	local bBarsDropDown = _G['bBarsDropDown'] or CreateFrame('Frame', 'bBarsDropDown', nil, 'UIDropDownMenuTemplate')
	UIDropDownMenu_Initialize(bBarsDropDown, function()
		local button = {
			{ text = 'Show keybinds', func = function(self) if bBarsDB[pname].keybindings == 1 then bBarsDB[pname].keybindings = 0 else bBarsDB[pname].keybindings = 1 end bBars_UpdateAll() end, checked = bBarsDB[pname].keybindings == 1, isNotRadio = true, keepShownOnClick = true, icon = 'Interface\\MacroFrame\\MacroFrame-Icon' },
			{ text = 'Show macro', func = function(self) if bBarsDB[pname].macroses == 1 then bBarsDB[pname].macroses = 0 else bBarsDB[pname].macroses = 1 end bBars_UpdateAll() end, checked = bBarsDB[pname].macroses == 1, isNotRadio = true, keepShownOnClick = true, icon = 'Interface\\MacroFrame\\MacroFrame-Icon' },
			{ text = 'Stick bars', func = function(self) if bBarsDB[pname].stick == 1 then bBarsDB[pname].stick = 0 else bBarsDB[pname].stick = 1 end end, checked = bBarsDB[pname].stick == 1, isNotRadio = true, keepShownOnClick = true },
			{ text = '|cffdd2200Reset settings|r', func = function(self) bBarsDB = nil ReloadUI() end, notCheckable = true },
		}
		 
		for i=1, #button do	UIDropDownMenu_AddButton(button[i])	end
	end, 'MENU')
	

	local MinimapButton = _G['bBarsMinimapButton'] or CreateFrame('Button', 'bBarsMinimapButton', Minimap)
	MinimapButton:EnableMouse(true)
	MinimapButton:SetMovable(true)
	MinimapButton:SetToplevel(true)
	MinimapButton:SetWidth(32)
	MinimapButton:SetHeight(32)
	MinimapButton:SetPoint('TOPLEFT')
	MinimapButton:SetFrameStrata('LOW')
	MinimapButton:RegisterForClicks('AnyUp')
	MinimapButton:RegisterForDrag('LeftButton')

	MinimapButton.border = MinimapButton.border or  MinimapButton:CreateTexture(nil, 'ARTWORK')
	MinimapButton.border:SetWidth(54)
	MinimapButton.border:SetHeight(54)
	MinimapButton.border:SetPoint('TOPLEFT')
	MinimapButton.border:SetTexture('Interface\\Minimap\\MiniMap-TrackingBorder')

	MinimapButton.bg = MinimapButton.bg or MinimapButton:CreateTexture(nil, 'BACKGROUND')
	MinimapButton.bg:SetWidth(24)
	MinimapButton.bg:SetHeight(24)
	MinimapButton.bg:SetPoint('TOPLEFT', 2, -4)
	MinimapButton.bg:SetTexture('Interface\\Minimap\\UI-Minimap-Background')

	MinimapButton.nt = MinimapButton.nt or MinimapButton:CreateTexture(nil, 'ARTWORK')
	MinimapButton.nt:SetWidth(12)
	MinimapButton.nt:SetHeight(14)
	MinimapButton.nt:SetPoint('TOPLEFT', 10.5, -8.5)
	MinimapButton.nt:SetTexture('Interface\\RAIDFRAME\\Raid-Icon-DebuffCurse')
	MinimapButton.nt:SetTexCoord(2/16, 14/16, 1/16, 15/16)
	MinimapButton:SetNormalTexture(NormalTexture)

	MinimapButton.pushed = MinimapButton.pushed or MinimapButton:CreateTexture(nil, 'ARTWORK')
	MinimapButton.pushed:SetWidth(10)
	MinimapButton.pushed:SetHeight(12)
	MinimapButton.pushed:SetPoint('TOPLEFT', 11.5, -9.5)
	MinimapButton.pushed:SetTexture('Interface\\RAIDFRAME\\Raid-Icon-DebuffCurse')
	MinimapButton.pushed:SetTexCoord(2/16, 14/16, 1/16, 15/16)
	MinimapButton:SetPushedTexture(MinimapButton.pushed)

	MinimapButton.highlight = MinimapButton.highlight or MinimapButton:CreateTexture(nil, 'ARTWORK')
	MinimapButton.highlight:SetPoint('TOPLEFT', 0, 0)
	MinimapButton.highlight:SetPoint('BOTTOMRIGHT', 0, 0)
	MinimapButton.highlight:SetTexture('Interface\\Minimap\\UI-Minimap-ZoomButton-Highlight')
	MinimapButton:SetHighlightTexture(MinimapButton.highlight)

	MinimapButton:SetScript('OnEnter', function(self)
		GameTooltip:SetOwner(self, 'ANCHOR_RIGHT')
		GameTooltip:AddLine('|cff00ff00bBars|r')
		GameTooltip:Show()
	end)
	
	MinimapButton:SetScript('OnLeave', function(self) GameTooltip:Hide() end)
	
	MinimapButton:SetScript('OnClick', function(self, button)
		self:UnlockHighlight()
		if IsShiftKeyDown() and button == 'LeftButton' then
			ReloadUI()
		elseif IsShiftKeyDown() and button == 'RightButton' then
			bBars_config = false
			bBars_ConfigBars(bBars_config)
			bBars_HoverBind()
		elseif not IsModifierKeyDown() then
			if (button == 'LeftButton') then
				bBars_config = not bBars_config
				bBars_ConfigBars(bBars_config)
			elseif (button == 'RightButton') then
				ToggleDropDownMenu(1, nil, bBarsDropDown, self, -0, -0)
				GameTooltip:SetOwner(self, 'ANCHOR_BOTTOMLEFT')
			end
		end
	end)
	
	MinimapButton:SetScript('OnDragStart', function(self) self:LockHighlight() self:SetScript('OnUpdate', DragMinimapButton) end)
	MinimapButton:SetScript('OnDragStop', function(self) self:SetScript('OnUpdate', nil) self:UnlockHighlight() end)
	MoveMinimapButton()
end

bBars:RegisterEvent('PLAYER_LEVEL_UP')
bBars:RegisterEvent('PLAYER_ENTERING_WORLD')
bBars:RegisterEvent('PLAYER_TALENT_UPDATE')
bBars:RegisterEvent('ADDON_LOADED')
bBars:SetScript('OnEvent', function(self, event, addon, ...)
	if (event == 'ADDON_LOADED' and addon == addonName) or event == 'PLAYER_ENTERING_WORLD' then
		if LBF then
			LBFGroup = LBFGroup or LBF:Group(addonName)
			LBFGroup:Skin(bBarsDB[pname].lbfSkin or 'Blizzard', bBarsDB[pname].lbfGloss or .4, bBarsDB[pname].lbfBackdrop, bBarsDB[pname].lbfColors)
			LBF:RegisterSkinCallback(addonName, function(a, SkinID, Gloss, Backdrop, Group, Button, Colors) 
				bBarsDB[pname].lbfSkin = SkinID
				bBarsDB[pname].lbfGloss = Gloss
				bBarsDB[pname].lbfBackdrop = Backdrop
				bBarsDB[pname].lbfColors = Colors
			end, '1')
		end

		if bBarsDB == nil then bBarsDB = {} end
		if bBarsDB[pname] == nil then bBarsDB[pname] = {} end
		if bBarsDB[pname].stick == nil then bBarsDB[pname].stick = 1 end
		if bBarsDB[pname].keybindings == nil then bBarsDB[pname].keybindings = 1 end
		if bBarsDB[pname].macroses == nil then bBarsDB[pname].macroses = 1 end
		if bBarsDB[pname].minimapPos == nil then bBarsDB[pname].minimapPos = 1 end
		if bBarsDB[pname].Minimap == nil then bBarsDB[pname].Minimap = 1 end

		bBars_HideBlizzard()
		bBars_MakeExpBarFrame()
		bBars_MakeRepBarFrame()
		for i = 1, 6 do bBars_MakeActionButtonsFrame('bMultiBar'..i, i) end
		if (class ~= 'MAGE' and class ~= 'SHAMAN') then bBars_MakeShapeShiftFrame() end
		if not IsPandaria then bBars_MakeTotemFrame() end
		bBars_MakeMicroMenuFrame()
		bBars_MakePetFrame()
		bBars_MakeBagsFrame()
		bBars_MakeVehicleFrame()
		bBars_MakeExtraBarFrame()
		bBars_MakePetBattleFrame()
		bBars_MakeBattlePetExpBarFrame()
		bBars_WatchFrame()

		bBars_TabBinder()
		bBars_OmniCC()
		bBars_CreateMinimapButton()
		bBars_ConfigBars(false)
		
		if bBarsDB[pname].Minimap == 1 then bBarsMinimapButton:Show() else bBarsMinimapButton:Hide() end
		self:UnregisterEvent('ADDON_LOADED')
	elseif event=='PLAYER_TALENT_UPDATE' or event == 'PLAYER_ENTERING_WORLD' then
		if bMultiBar1 then
			bMultiBar1:SetAttribute('_onstate-page', [[ control:ChildUpdate('page', newstate) ]])
			RegisterStateDriver(bMultiBar1, 'page', bBars_getBar())
		end
		bBars_UpdateAll()
		collectgarbage()
	end
end)

-- setup slash command
SLASH_Bars1 = '/bbars';
SLASH_Bars2 = '/bb';
SlashCmdList['Bars'] = function(cmd)
	cmd = lower(cmd)
	if InCombatLockdown() or UnitAffectingCombat('player') then return end
	if (cmd == 'reset') then
		bBarsDB = {}
		ReloadUI()
	elseif (cmd == 'minimap') then
		bBarsDB[pname].Minimap = (bBarsDB[pname].Minimap == 1 and 0 or 1)
		if bBarsDB[pname].Minimap == 1 then _G['bBarsMinimapButton']:Show() else _G['bBarsMinimapButton']:Hide() end
		print('|cff32a1e1bBars:|r|cffffffffMinimap button '..(bBarsDB[pname].Minimap == 1 and 'on' or 'off')..'|r')
	elseif (cmd == 'bind') then
		bBars_config = false
		bBars_ConfigBars(false)
		bBars_HoverBind()
	else
		bBars_config = not bBars_config
		bBars_ConfigBars(bBars_config)
		print('|cff32a1e1bBars:|r|cffffffffConfig mode '..(bBars_config and'on' or 'off')..'|r')
	end
end